// File:        LightVis.cpp 
// Copyright:   (c) 2009 by Frydaysoft.de
// Author:      SFr (frydaysoft@gmx.de)
// Proj:        light visualization plugin (winamp)
// Short Info:  Basic application and utility functions of the plugin

// ***************************************************************************
// * Winamp visualization plugin
// * This plugin analyzes the audio data and generates light pattern.
// * These pattern will be displayed on screen and transmitted to the 
// * LightUnit over the selected COM port 
// ***************************************************************************

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <time.h>
#include "winamp\wa_ipc.h"
#include "sio\Win32Serial.h"
#include "vis.h"
#include "resource.h"


// ***************************************************************************
// ***** convert given (unicode) string into a double value 
#ifndef _ttof
  #ifndef _UNICODE
    // ***** use function from C-lib
    #define _ttof atof
  #else
    // ***** not part of the C-lib (only: _ttoi/_ttol)
    double _ttof(TCHAR *p)
    {
      #error "implementation tbd!"
      // e.g: convert to ansi and use atof()
    }
  #endif
#endif

// ***************************************************************************
#define DEB_DEF                   3
#define DEB_MIN                   1
#define DEB_MAX                   6
#define DEB_WND_TEXT              "debounce (1..6)"
#define RETRIG_DEF                90
#define RETRIG_MIN                60
#define RETRIG_MAX                3000
#define RETRIG_WND_TEXT           "retrigger (60..3000ms)"
#define CMP_DIFF_SPECTR_DEF       35 
#define CMP_DIFF_SPECTR_MIN       2
#define CMP_DIFF_SPECTR_MAX       120
#define CMP_DIFF_SPECTR_WND_TEXT "cmp_diff (2..120)"
#define CMP_DIFF_VU_DEF           80
#define CMP_DIFF_VU_MIN           10
#define CMP_DIFF_VU_MAX           200
#define CMP_DIFF_VU_WND_TEXT      "cmp_diff (10..200)"

#define CHANGE_DEF        2   // nr of bits that can be changed from old to new lights value
#define CHANGE_MIN        0
#define CHANGE_MAX        4
#define LIGHTS_OFF_DEF    2   // nr of bits that must be ON in new lights value
#define LIGHTS_OFF_MIN    0
#define LIGHTS_OFF_MAX    5
#define LIGHTS_ON_DEF     5   // maximum nr of bits that can be ON in new lights value
#define LIGHTS_ON_MIN     0
#define LIGHTS_ON_MAX     8
// defaults see DefaultReduceSpectr[]
#define REDUCE_MIN        0.01  // allowed min
#define REDUCE_MAX        2.00  // allowed max
double DefaultReduceSpectr[8] = 
    // 0     1     2     3     4     5     6     7
    {0.10, 0.31, 0.62, 0.80, 0.90, 0.95, 0.90, 0.90};

// ***************************************************************************
// GUI IDs to access these controls
int UID_ReduceCtrl[8]  = { IDC_EDIT_RED8, IDC_EDIT_RED7, IDC_EDIT_RED6, IDC_EDIT_RED5,
                           IDC_EDIT_RED4, IDC_EDIT_RED3, IDC_EDIT_RED2, IDC_EDIT_RED1};
int UID_VuMeter[]      = { IDC_CHECK_USE_WAVE, IDC_EDIT_CH_DIFF, IDC_EDIT_MIN, IDC_EDIT_MAX};
int UID_SpecAnalyzer[] = { IDC_EDIT_RED8, IDC_EDIT_RED7, IDC_EDIT_RED6, IDC_EDIT_RED5,
                           IDC_EDIT_RED4, IDC_EDIT_RED3, IDC_EDIT_RED2, IDC_EDIT_RED1};


// ***************************************************************************
// some constant paint position for the VU meter
#define VU_MOD_X_SIZE   256
#define VU_LAMP_SIZE    (VU_MOD_X_SIZE/8)
#define VU_LAMP_HEIGHT  (VU_LAMP_SIZE+2)
#define VU_HEIGHT       16
#define VU_INFO_HEIGHT  32
#define VU_MOD_Y_SIZE   (VU_LAMP_HEIGHT + 2*VU_HEIGHT + VU_INFO_HEIGHT)
#define VU_LAMP_Y_POS   (0)
#define VU_A_Y_POS      (VU_LAMP_HEIGHT)
#define VU_B_Y_POS      (VU_LAMP_HEIGHT+VU_HEIGHT)
#define VU_INFO_Y_POS   (VU_LAMP_HEIGHT + 2*VU_HEIGHT)
// some constant paint position for the spectrum analyzer
#define SPEC_MOD_X_SIZE   256
#define SPEC_LAMP_SIZE    (SPEC_MOD_X_SIZE/8)
#define SPEC_LAMP_HEIGHT  (SPEC_LAMP_SIZE+2)
#define SPEC_HEIGHT       (288/2)
#define SPEC_INFO_HEIGHT  32
#define SPEC_MOD_Y_SIZE   (SPEC_LAMP_HEIGHT + SPEC_HEIGHT + SPEC_INFO_HEIGHT)
#define SPEC_LAMP_Y_POS   (0)
#define SPEC_Y_POS        (SPEC_LAMP_HEIGHT)
#define SPEC_INFO_Y_POS   (SPEC_LAMP_HEIGHT + SPEC_HEIGHT)


// ***************************************************************************
// text array with valid COM ports (for this PC)
#define TXT_COM_LEN 7                                 // "COM 256"
TCHAR *RealComPorts[MAX_NUM_COMPORTS][TXT_COM_LEN+1];
RS232_config g_SioCfg;

// ***************************************************************************
// "utility" functions
int  ScanComPorts(void);
void config_read(struct winampVisModule *this_mod);		                  // reads the configuration
void config_write(struct winampVisModule *this_mod, bool bPos);	        // writes the configuration
void config_getinifn(struct winampVisModule *this_mod, TCHAR *ini_file); // makes the .ini file filename
winampVisModule *getModule(int which);                                  // returns a winampVisModule when requested. Used in hdr, below

// ***************************************************************************
// "member" functions
void config(struct winampVisModule *this_mod);                          // configuration dialog
int  init(struct winampVisModule *this_mod);	                          // initialization for module
int  render1(struct winampVisModule *this_mod);                         // rendering for module 1
int  render2(struct winampVisModule *this_mod);                         // rendering for module 2
void quit(struct winampVisModule *this_mod);                            // deinitialization for module
// our procedure for the application window
LRESULT CALLBACK WndAppProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
// our procedure for the configuration window
LRESULT CALLBACK WndConfigProc(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam);

// ***************************************************************************
TCHAR szAppName[] = "LightVisualization";                               // Our window class, etc

embedWindowState  g_myWindowState;
winampVisModule   *g_app_plugin = NULL, *g_cfg_plugin = NULL;
int               g_width, g_height;
bool              g_random_presets_flag = 0;
HWND              g_hMainWnd;                         // main window handle

// ***************************************************************************
// configuration declarations
struct
{
  bool            CfgIsValid;                         // true if config loaded from INI-File
  int             config_x, config_y;	                // screen X position and Y position, repsectively
  int             ComPort;                            // no of com port
  int             DebounceCnt;                        // nr of dbounce loops
  DWORD           RefreshTime;                        // time slice for value uptate
  int             CmpDiffSpectr;                      // compare diff. for spectrum alalyzer
  int             CmpDiffVu;                          // compare diff. for VU-meter
  bool            VuUsesWaveData;                     // 0: use spectr. data, 1: waveform data
  BYTE            ChangeCnt;                          // nr of bits allowed for changing on update of light value
  BYTE            MinCnt;                             // minimum of bits that must be ON
  BYTE            MaxCnt;                             // maximum of bits that can be ON
  double          ReduceSpec[8];                      // reducer multiplier for spectrum analyzer
}
cfg_change,                                           // temp. cfg in CFG dialog
cfg =                                                 // active cfg 
{ 
    false,                                              // CfgIsValid
    50, 50,                                             // config_x, config_y
    1,                                                  // ComPort
    DEB_DEF,                                            // DebounceCnt
    RETRIG_DEF,                                         // RetriggerSlot
    CMP_DIFF_SPECTR_DEF,                                // CmpDiffSpectr
    CMP_DIFF_VU_DEF,                                    // CmpDiffVu
    false,                                              // VuUsesWaveData
    CHANGE_DEF, LIGHTS_OFF_DEF, LIGHTS_ON_DEF,          // ChangeCnt, MinCnt, MaxCnt
    {DefaultReduceSpectr[0], 
     DefaultReduceSpectr[1], 
     DefaultReduceSpectr[2], 
     DefaultReduceSpectr[3], 
     DefaultReduceSpectr[4], 
     DefaultReduceSpectr[5], 
     DefaultReduceSpectr[6], 
     DefaultReduceSpectr[7]}                            // ReduceSpec
};

// ***************************************************************************
#define TX_MAX_LEN  6     // ">3L??\n" or ">2JC\n"
struct
{
  HANDLE          hCommPort;
  DWORD           LastTickCnt;                        // for time calculation
  DWORD           LastTickCmp;                        // compare time to detect next time slice for update of light value
  int             LastCmpPeak[8];                     // 1 for VU, 8 for spectrum
  BYTE            Lights;                             // current value for LightUnit
  char            TxBuf[TX_MAX_LEN+1];                // tx buffer
  int             PeakHistory[8][DEB_MAX];            // 2 for VU, 8 for spectrum
  int             PeakIdx;                            // index into PeakHistory[]
  bool            PeakValid;                          // true if PeakHistory[] filled
}lights = 
{
            INVALID_HANDLE_VALUE,                     // hCommPort
            0,                                        // LastTickCnt
            0,                                        // LastTickCmp
            0,                                        // LastCmpPeak
            0,                                        // Lights
            {0},                                      // TxBuf[]
            {0},                                      // PeakHistory[]
            0,                                        // PeakIdx
            0,                                        // PeakValid
};


// ***************************************************************************
// Double buffering data
HDC g_memDC;		        // memory device context
HBITMAP	g_memBM,        // memory bitmap (for g_memDC)
		    g_oldBM;        // old bitmap (from g_memDC)
HBRUSH  g_red_brush;
HBRUSH  g_green_brush;
HBRUSH  g_yellow_brush;

// ***************************************************************************
// Module header, includes version, description, and address of the module retriever function
winampVisHeader hdr = { VIS_HDRVER, "Light Visualisation v1.0 (www.frydaysoft.de)", getModule };


// ***************************************************************************
// module: spectrum analyzer
winampVisModule mod_spectr =
{
	"Light Spectrum Analyzer",
	NULL,	    // hwndParent
	NULL,	    // hDllInstance
	0,		    // sRate
	0,		    // nCh
	25,		    // latencyMS
	25,		    // delayMS
	2,		    // spectrumNch
	0,		    // waveformNch
	{ 0, },	  // spectrumData
	{ 0, },	  // waveformData
	config,
	init,
	render1, 
	quit
};

// ***************************************************************************
// module: VU meter
winampVisModule mod_vu =
{
	"Light VU Meter",
	NULL,	    // hwndParent
	NULL,	    // hDllInstance
	0,		    // sRate
	0,		    // nCh
	25,		    // latencyMS
	25,		    // delayMS
	2,		    // spectrumNch
	2,		    // waveformNch
	{ 0, },	  // spectrumData
	{ 0, },	  // waveformData
	config,
	init,
	render2, 
	quit
};


// ***************************************************************************
// this is the only exported symbol. returns our main header.
// if you are compiling C++, the extern "C" { is necessary, so we just #ifdef it
#ifdef __cplusplus
extern "C" {
#endif
__declspec( dllexport ) winampVisHeader *winampVisGetHeader()
{
	return &hdr;
}
#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  // fdwReason
  // 1: start/init
  // 2: processing
  // 3: stop/end
	return TRUE;
}
#ifdef __cplusplus
}
#endif



// ***************************************************************************
// getmodule routine from the main header. Returns NULL if an invalid module was requested,
// otherwise returns either mod_spectr or mod_vu depending on 'which'.
winampVisModule *getModule(int which)
{
	switch (which)
	{
		case 0: return &mod_spectr;
		case 1: return &mod_vu;
		default:return NULL;
	}
}


//******************************************************************************
// scans physical COM ports, existing ports will be noticed into string array
int ScanComPorts(void)
{
	TCHAR ComString[7];              // for max: "COM999"
  int   cnt = 0;

  memset(RealComPorts, 0, sizeof(RealComPorts));

  for (int i=1; i<MAX_NUM_COMPORTS; i++)
  {
#pragma warning(disable:4996)   
	  _stprintf(ComString, "COM%d", i);
#pragma warning(default:4996)   

    BOOL bSuccess = FALSE;
    HANDLE hPort = ::CreateFile(ComString, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
    if (hPort == INVALID_HANDLE_VALUE)
    {
      DWORD dwError = GetLastError();
      if (dwError == ERROR_ACCESS_DENIED || dwError == ERROR_GEN_FAILURE || dwError == ERROR_SHARING_VIOLATION || dwError == ERROR_SEM_TIMEOUT)
        bSuccess = TRUE;
    }
    else
    {
      bSuccess = TRUE;
      CloseHandle(hPort);
    }

    if (bSuccess)
    {
#pragma warning(disable:4996) 
      _stprintf((TCHAR*)&RealComPorts[cnt], "COM %d", i);
#pragma warning(default:4996) 
      cnt++;
    }
  }

  return cnt;
}


// ***************************************************************************
LRESULT CALLBACK WndConfigProc(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  static TCHAR Info[50], ComStr[10];
  TCHAR *pBuf;
  HWND ComboHwnd, Tmp;
  int i, PortSel, Val;

	switch (uMsg)
	{
		case WM_INITDIALOG:
			{
        if ( !cfg.CfgIsValid )
        { // ***** load settings from INI file
          config_read(g_cfg_plugin);
        }
        cfg_change = cfg;

        SetWindowText(hwndDlg, "config dialog");

        // ***** COM port
        ComboHwnd = GetDlgItem(hwndDlg, IDC_COMBO_SEL);
        if ( ComboHwnd )
        {
          PortSel = 0;
          if ( lights.hCommPort != INVALID_HANDLE_VALUE )
          {
#pragma warning(disable:4996) 
            _stprintf(ComStr, "Com%d", g_SioCfg.ComPort);
#pragma warning(default:4996) 
            SendMessage(ComboHwnd, CB_ADDSTRING, (WPARAM)0, (LPARAM)&ComStr[0]);
			::EnableWindow(ComboHwnd, false);
            PortSel = 0;
          }
          else
          {
				    for (i = 0; i < MAX_NUM_COMPORTS; i++)
            {
              pBuf = (TCHAR*)&RealComPorts[i][0];
              if ( *pBuf )
              {
                // ***** use sscanf not in a dll !!!
                Val = _ttoi(&pBuf[sizeof("COM")]);
                if ( Val == cfg.ComPort )
                {
                  PortSel = i;
                }
                SendMessage(ComboHwnd, CB_ADDSTRING, (WPARAM)0, (LPARAM)&RealComPorts[i]);
              }
              else
              {
                break;
              }
            }
          }
        }
        SendMessage(ComboHwnd, CB_SETCURSEL, (WPARAM)PortSel, (LPARAM)0);

        // ***** common configuration
        SetDlgItemText(hwndDlg,IDC_TEXT_DEB,      DEB_WND_TEXT);
        SetDlgItemInt(hwndDlg,IDC_EDIT_DEB,       cfg.DebounceCnt, FALSE);

        SetDlgItemText(hwndDlg,IDC_TEXT_RETRIG,   RETRIG_WND_TEXT);
        SetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG,    cfg.RefreshTime, FALSE);
        if (g_cfg_plugin == &mod_spectr)
        {
          SetDlgItemText(hwndDlg,IDC_TEXT_CMP_DIFF, CMP_DIFF_SPECTR_WND_TEXT);
          SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  cfg.CmpDiffSpectr, FALSE);
				  for (i = 0; i < sizeof(UID_VuMeter)/sizeof(int); i++)
          {
            Tmp = GetDlgItem(hwndDlg, UID_VuMeter[i]);
			::EnableWindow(Tmp, FALSE);
          }
        }
        if (g_cfg_plugin == &mod_vu)
        {
          SetDlgItemText(hwndDlg,IDC_TEXT_CMP_DIFF, CMP_DIFF_VU_WND_TEXT);
          SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  cfg.CmpDiffVu, FALSE);
				  for (i = 0; i < sizeof(UID_SpecAnalyzer)/sizeof(int); i++)
          {
            Tmp = GetDlgItem(hwndDlg, UID_SpecAnalyzer[i]);
			::EnableWindow(Tmp, FALSE);
          }
        }

        // ***** spectrum analyzer
				for (i = 0; i < sizeof(UID_ReduceCtrl)/sizeof(int); i++)
        {
#pragma warning(disable:4996) 
          _stprintf(Info, "%1.2lf", cfg.ReduceSpec[i]);
#pragma warning(default:4996) 
          SetDlgItemText(hwndDlg, UID_ReduceCtrl[i], Info);
        }

        // ***** VU-meter
			  CheckDlgButton(hwndDlg,IDC_CHECK_USE_WAVE, cfg.VuUsesWaveData?BST_CHECKED:BST_UNCHECKED);
        SetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF,  cfg.ChangeCnt, FALSE);
        SetDlgItemInt(hwndDlg,IDC_EDIT_MIN,  cfg.MinCnt, FALSE);
        SetDlgItemInt(hwndDlg,IDC_EDIT_MAX,  cfg.MaxCnt, FALSE);

        // ***** status line
        if (g_cfg_plugin == &mod_spectr)
        {
          SetDlgItemText(hwndDlg,IDC_STATUS, "mod_spectr");
        }
        if (g_cfg_plugin == &mod_vu)
        {
          SetDlgItemText(hwndDlg,IDC_STATUS, "mod_vu");
        }
			}
		break;
		case WM_COMMAND:
      switch (LOWORD(wParam))
      {
        case IDC_BTN_LD_DEF:  //1002
          // ***** relaod defaults
          SetDlgItemInt(hwndDlg,IDC_EDIT_DEB,       DEB_DEF, FALSE);
          SetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG,    RETRIG_DEF, FALSE);
          if (g_cfg_plugin == &mod_spectr)
          {
            SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  CMP_DIFF_SPECTR_DEF, FALSE);
          }
          if (g_cfg_plugin == &mod_vu)
          {
            SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  CMP_DIFF_VU_DEF, FALSE);
          }
          // ***** spectrum analyzer
				  for (i = 0; i < sizeof(UID_ReduceCtrl)/sizeof(int); i++)
          {
#pragma warning(disable:4996) 
            _stprintf((TCHAR*)&Info, "%1.2lf", DefaultReduceSpectr[i]);
#pragma warning(default:4996) 
            SetDlgItemText(hwndDlg, UID_ReduceCtrl[i], Info);
          }

          // ***** VU-meter
  			  CheckDlgButton(hwndDlg,IDC_CHECK_USE_WAVE, BST_UNCHECKED);
          SetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF,  CHANGE_DEF, FALSE);
          SetDlgItemInt(hwndDlg,IDC_EDIT_MIN,  LIGHTS_OFF_DEF, FALSE);
          SetDlgItemInt(hwndDlg,IDC_EDIT_MAX,  LIGHTS_ON_DEF, FALSE);
          break;
        case IDCANCEL:        // 2
          EndDialog(hwndDlg,0);
          break;
        case IDOK:            // 1
          ComboHwnd = GetDlgItem(hwndDlg, IDC_COMBO_SEL);
          if ( ComboHwnd )
          {
            // ***** COM port
            if ( IsWindowEnabled(ComboHwnd) )
            {
              i = SendMessage(ComboHwnd, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
#pragma warning(disable:4996) 
              _stprintf((TCHAR*)&Info, "sel: %d", i);
#pragma warning(default:4996) 
              i = SendMessage(ComboHwnd, CB_GETLBTEXT, (WPARAM)i, (LPARAM)Info);
              if ( i != CB_ERR )
              {
                // ***** use sscanf not in a dll !!!
                Val = _ttoi((TCHAR*)&Info[sizeof("COM")]);
                cfg_change.ComPort = Val;                
              }
            }
            cfg_change.DebounceCnt = GetDlgItemInt(hwndDlg,IDC_EDIT_DEB, NULL, FALSE);
            if ( (cfg_change.DebounceCnt > DEB_MAX) || (cfg_change.DebounceCnt < DEB_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"debounce out of range"); break; }
            cfg_change.RefreshTime = GetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG, NULL, FALSE);
            if ( (cfg_change.RefreshTime > RETRIG_MAX) || (cfg_change.RefreshTime < RETRIG_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"retrigger out of range"); break; }
            if (g_cfg_plugin == &mod_spectr)
            {
              cfg_change.CmpDiffSpectr = GetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF, NULL, FALSE);
              if ( (cfg_change.CmpDiffSpectr > CMP_DIFF_SPECTR_MAX) || (cfg_change.CmpDiffSpectr < CMP_DIFF_SPECTR_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"cmp_diff out of range"); break; }
            }
            if (g_cfg_plugin == &mod_vu)
            {
              cfg_change.CmpDiffVu = GetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF, NULL, FALSE);
              if ( (cfg_change.CmpDiffVu > CMP_DIFF_VU_MAX) || (cfg_change.CmpDiffVu < CMP_DIFF_VU_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"cmp_diff out of range"); break; }
            }

            // ***** spectrum analyzer
				    for (i = 0; i < sizeof(UID_ReduceCtrl)/sizeof(int); i++)
            {
              GetDlgItemText(hwndDlg, UID_ReduceCtrl[i], Info, sizeof(Info)/sizeof(TCHAR));
              // ***** use sscanf not in a dll !!!
              cfg_change.ReduceSpec[i] = _ttof(Info);
              if ( (cfg_change.ReduceSpec[i] > REDUCE_MAX) || (cfg_change.ReduceSpec[i] < REDUCE_MIN) )
              { 
#pragma warning(disable:4996) 
				  _stprintf(Info,"reduce beam(%d, %1.2f) out of range(%1.2f..%1.2f)",8-i,cfg_change.ReduceSpec[i],REDUCE_MIN,REDUCE_MAX); 
#pragma warning(default:4996) 
				  SetDlgItemText(hwndDlg,IDC_STATUS,Info); break; 
			  }
            }
            if (i != (sizeof(UID_ReduceCtrl)/sizeof(int)) )
            {
              break;
            }

            // ***** VU-meter
			      if ( IsDlgButtonChecked(hwndDlg,IDC_CHECK_USE_WAVE) == BST_CHECKED )
            { cfg_change.VuUsesWaveData = true; } else { cfg_change.VuUsesWaveData = false; }
            cfg_change.ChangeCnt = GetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF,  NULL, FALSE);
            if ( (cfg_change.ChangeCnt > CHANGE_MAX) || (cfg_change.ChangeCnt < CHANGE_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"change_diff of range"); break; }
            cfg_change.MinCnt    = GetDlgItemInt(hwndDlg,IDC_EDIT_MIN,  NULL, FALSE);
            if ( (cfg_change.MinCnt > LIGHTS_OFF_MAX) || (cfg_change.MinCnt < LIGHTS_OFF_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"min_lights of range"); break; }
            cfg_change.MaxCnt    = GetDlgItemInt(hwndDlg,IDC_EDIT_MAX,  NULL, FALSE);
            if ( (cfg_change.MaxCnt > LIGHTS_ON_MAX) || (cfg_change.MaxCnt < LIGHTS_ON_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"max_lights of range"); break; }
            if ( (cfg_change.MinCnt > cfg_change.MaxCnt ) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"min_lights > max_lights"); break; }
          }
          // ***** store settings
          cfg = cfg_change;
          config_write(g_cfg_plugin, false); // write cfg part
          EndDialog(hwndDlg,0);
          break;
        default:
          break;
      }
    break;
	}
	return 0;
}


// ***************************************************************************
void config(struct winampVisModule *this_mod)
{
  g_cfg_plugin = this_mod;
  ScanComPorts();
  DialogBox(g_cfg_plugin->hDllInstance,MAKEINTRESOURCE(IDD_DIALOG1),g_cfg_plugin->hwndParent, (DLGPROC)WndConfigProc);
}


// ***************************************************************************
// initialization. Registers our window class, creates our window, etc. 
// Note: this one works for both modules, but you should make init1() and init2()...
// returns 0 on success, 1 on failure.
int init(struct winampVisModule *this_mod)
{
  int styles;
  HWND parent = NULL;
  HWND (*e)(embedWindowState *v);

  // ***** set some global utility data members
  g_app_plugin = this_mod;
	g_width = (this_mod == &mod_vu)?VU_MOD_X_SIZE:SPEC_MOD_X_SIZE;
	g_height = (this_mod == &mod_vu)?(VU_MOD_Y_SIZE):SPEC_MOD_Y_SIZE;

  // ***** basic data initialization
  lights.LastTickCmp = GetTickCount();
  lights.LastTickCnt = GetTickCount();
  memset(&lights.LastCmpPeak, 0, sizeof(lights.LastCmpPeak));;
  srand( (unsigned)time( NULL ) );
  lights.Lights      = 0;
  lights.PeakIdx     = 0;
  lights.PeakValid   = 0;

	config_read(this_mod);

  g_myWindowState.flags = 0;
  // uncomment this line if your plugin draws to the screen using directx OVERLAY mode
  // g_myWindowState.flags |= EMBED_FLAGS_NOTRANSPARENCY; 
  g_myWindowState.r.left    = cfg.config_x;
  g_myWindowState.r.top     = cfg.config_y;

  g_myWindowState.r.right   = cfg.config_x + g_width;
  if (this_mod == &mod_spectr)
  {
    // ***** note: winamp resizes the window in tabbed steps so set the height 
    //             a little bit greather for the spectrum analyzer
    g_myWindowState.r.bottom  = cfg.config_y + g_height + 50;
  }
  else
  {
    g_myWindowState.r.bottom  = cfg.config_y + g_height;
  }
     
  *(void**)&e = (void *)SendMessage(this_mod->hwndParent,WM_WA_IPC,(LPARAM)0,IPC_GET_EMBEDIF);

  if (!e)
  {
		MessageBox(this_mod->hwndParent,"This plugin requires Winamp 5.0+","error",MB_OK);
		return 1;
  }

  parent = e(&g_myWindowState);

  SetWindowText(g_myWindowState.me, this_mod->description); // set our window title
	
  // ***** Register our window class
	{	
		WNDCLASS wc;
		memset(&wc,0,sizeof(wc));
		wc.lpfnWndProc = WndAppProc;			        // our window procedure
		wc.hInstance = this_mod->hDllInstance;	  // hInstance of DLL
		wc.lpszClassName = szAppName;			        // our window class name
		if (!RegisterClass(&wc)) 
		{
			MessageBox(this_mod->hwndParent,"Error registering window class","error",MB_OK);
			return 1;
		}
	}

  // ***** create out application window
  styles = WS_VISIBLE|WS_CHILDWINDOW|WS_OVERLAPPED|WS_CLIPCHILDREN|WS_CLIPSIBLINGS;
	g_hMainWnd = CreateWindowEx(
		0,	                      // these exstyles put a nice small frame, 
											        // but also a button in the taskbar
		szAppName,							  // our window class name
		NULL,				              // no title, we're a child
		styles,				            // do not make the window visible 
		cfg.config_x,cfg.config_y,// screen position (read from config)
    g_width, g_height,			  // width & height of window (need to adjust client area later)
		parent,				            // parent window (winamp main window)
		NULL,								      // no menu
		this_mod->hDllInstance,	  // hInstance of DLL
		0);                       // no window creation data

	if (!g_hMainWnd) 
	{
		MessageBox(this_mod->hwndParent,"Error creating window","error",MB_OK);
		return 1;
	}

	SetWindowLong(g_hMainWnd,GWL_USERDATA,(LONG)this_mod); // set our reference to a "this" pointer
  SendMessage(this_mod->hwndParent, WM_WA_IPC, (int)g_hMainWnd, IPC_SETVISWND);

	// ***** create our doublebuffer for paint()
  BITMAPINFO bmi;
  g_memDC = CreateCompatibleDC(NULL);
  ZeroMemory(&bmi, sizeof bmi);
  bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi.bmiHeader.biWidth = g_width;
  bmi.bmiHeader.biHeight = -g_height;
  bmi.bmiHeader.biPlanes = 1;
  bmi.bmiHeader.biBitCount = 32;
  bmi.bmiHeader.biCompression = BI_RGB;
  bmi.bmiHeader.biSizeImage = 0;
  bmi.bmiHeader.biXPelsPerMeter = 0;
  bmi.bmiHeader.biYPelsPerMeter = 0;
  bmi.bmiHeader.biClrUsed = 0;
  bmi.bmiHeader.biClrImportant = 0;
  void *bits = 0;
  g_memBM = CreateDIBSection(g_memDC, &bmi, DIB_RGB_COLORS, &bits, NULL, 0);
  g_oldBM = (HBITMAP)SelectObject(g_memDC,g_memBM);

  // ***** some static brushes for faster painting
  g_red_brush   = CreateSolidBrush( RGB(255,0,0) ); 
  g_green_brush = CreateSolidBrush( RGB(0,255,0) ); 
  g_yellow_brush= CreateSolidBrush( RGB(255,255,0) ); 

  // ***** clear our screen with a white color 
  {
    RECT r={0, 0, g_width, g_height};
    FillRect(g_memDC, &r, (HBRUSH)GetStockObject(WHITE_BRUSH));
  }

  if ( InitComPortConfig(g_SioCfg, cfg.ComPort, CBR_57600) )
  {
    OpenComPort(lights.hCommPort, g_SioCfg);
    if ( lights.hCommPort != INVALID_HANDLE_VALUE )
    { // ***** send JumpToConfig --> the HandHelt can be used for testing also
#pragma warning(disable:4996)
		sprintf(lights.TxBuf, ">2JC\n");
#pragma warning(default:4996)
	  WriteComPort(lights.hCommPort, lights.TxBuf, 5); 
    }
  }

	// ***** show the window
	ShowWindow(parent,SW_SHOWNORMAL);
	return 0;
}


// ***************************************************************************
// render function for analyzer. 
// Returns 0 if successful, 1 if visualization should end.
int render1(struct winampVisModule *this_mod)
{
  TCHAR Info[20];
  DWORD TickCnt = GetTickCount();
  bool UpdateLigths = false;
  HBRUSH old = (HBRUSH)GetCurrentObject(g_memDC, OBJ_BRUSH);
	int x, msk;
	// ***** pre-calculate analyzer data above half spectrum data
  int PeakDiff, Peak, Idx, SpecColumn[8] = {0,0,0,0,0,0,0,0};
	for (x = 0; x < (576/2); x ++)
	{
    Idx = x / ((576/2)/8);        // combine data for upto 8 lights
    Idx &= 7;
    if ( this_mod->nCh > 1 )
    { // ***** use data from both channels
      SpecColumn[Idx] += (this_mod->spectrumData[0][x] + this_mod->spectrumData[1][x]) / 2;
    }
    else
    { // ***** use data from mono channel
      SpecColumn[Idx] += this_mod->spectrumData[0][x];
    }
	}

  // ***** clear background of spectrum view
	Rectangle(g_memDC, 0, SPEC_Y_POS, SPEC_MOD_X_SIZE, SPEC_Y_POS+SPEC_HEIGHT);

	// ***** pre-calculate analyzer data
  SelectObject(g_memDC, g_green_brush);
  for (x = 0, msk=128; x < 8; x ++, msk/=2)
  { // ***** trim pre-calculateted data into view space
    SpecColumn[x] = (int)((double)(SpecColumn[x]/2) * cfg.ReduceSpec[x]);
    if (SpecColumn[x] > (SPEC_HEIGHT) )
    {
      SpecColumn[x] = (SPEC_HEIGHT);
    }
    lights.PeakHistory[x][lights.PeakIdx] = SpecColumn[x];
    if ( ++lights.PeakIdx >= cfg.DebounceCnt )
    {
      lights.PeakIdx   = 0;
      lights.PeakValid = true;
    }
    // ***** draw (valid) debounced spectrum data
    if ( lights.PeakValid )
    {
      Peak = 0;
      for (Idx = 0; Idx < cfg.DebounceCnt; Idx ++)
      {
        Peak += lights.PeakHistory[x][Idx];
      }
      Peak /= cfg.DebounceCnt;
   	  Rectangle(g_memDC,x*(SPEC_MOD_X_SIZE/8),(SPEC_Y_POS+SPEC_HEIGHT)-Peak,(1+x)*(SPEC_MOD_X_SIZE/8),(SPEC_Y_POS+SPEC_HEIGHT));

      if ( (TickCnt-lights.LastTickCmp) >= cfg.RefreshTime )
      {
        if ( Peak < lights.LastCmpPeak[x] )
        {
          PeakDiff = -(lights.LastCmpPeak[x] - Peak);
        }
        else
        {
          PeakDiff = Peak - lights.LastCmpPeak[x];
        }

        if ( abs(PeakDiff) >= cfg.CmpDiffSpectr )
        {
          lights.LastCmpPeak[x] = Peak;
          if ( PeakDiff > 0 )
          { // ***** try to set ON
            if ( !(lights.Lights & msk) )
            {
              lights.Lights |= msk;
              UpdateLigths = true;
            }
          }
          else
          { // ***** try to set OFF
            if ( lights.Lights & msk )
            { 
              lights.Lights &= ((~(msk))&0xFF);
              UpdateLigths = true;
            }
          }
        }        
      }
    }
  }
  SelectObject(g_memDC, old);

  // ***** time for new update
  if ( UpdateLigths )
  {
    lights.LastTickCmp = TickCnt;
    // ***** clear background
    SelectObject(g_memDC, old);
	  Rectangle(g_memDC, 0, 0, SPEC_MOD_X_SIZE, SPEC_LAMP_HEIGHT);

    // ***** send to COM port
    if ( lights.hCommPort != INVALID_HANDLE_VALUE )
    {
#pragma warning(disable:4996)
		sprintf(lights.TxBuf, ">3L%2.2LX\n", lights.Lights);
#pragma warning(default:4996)
	  WriteComPort(lights.hCommPort, lights.TxBuf, 6); 
    }

    // ***** draw new value on screen
    SelectObject(g_memDC, g_yellow_brush);
    if ( lights.Lights & 128 ) Ellipse(g_memDC, 0*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (0+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 64  ) Ellipse(g_memDC, 1*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (1+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 32  ) Ellipse(g_memDC, 2*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (2+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 16  ) Ellipse(g_memDC, 3*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (3+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 8   ) Ellipse(g_memDC, 4*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (4+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 4   ) Ellipse(g_memDC, 5*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (5+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 2   ) Ellipse(g_memDC, 6*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (6+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
    if ( lights.Lights & 1   ) Ellipse(g_memDC, 7*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (7+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
  }

  SelectObject(g_memDC, old);
  // ***** append some space for erase backgrond of older (longer) text
  if ( lights.hCommPort != INVALID_HANDLE_VALUE )
  {
#pragma warning(disable:4996) 
    x = _stprintf(Info,"Refresh: %d ms    ", TickCnt - lights.LastTickCnt);
#pragma warning(default:4996) 
  }
  else
  {
#pragma warning(disable:4996) 
    x = _stprintf(Info,"no COM port !    ");
#pragma warning(default:4996) 
  }
  TextOut(g_memDC, 0, SPEC_INFO_Y_POS, Info, x);
  lights.LastTickCnt = TickCnt;

 	{ // copy doublebuffer to window
		HDC hdc = GetDC(g_hMainWnd);
		BitBlt(hdc, 0, 0, SPEC_MOD_X_SIZE, SPEC_MOD_Y_SIZE, g_memDC, 0, 0, SRCCOPY);
		ReleaseDC(g_hMainWnd,hdc);
	}
	return 0;
}


// ***************************************************************************
// render function for VU meter. 
// Returns 0 if successful, 1 if visualization should end.
int render2(struct winampVisModule *this_mod)
{
  TCHAR Info[20];
  bool UpdateLigths = false;
  int PeakDiff = 0;
  int curr_peak = 0;
  int total, LoopTimeout;
	int x, len;
  DWORD TickCnt = GetTickCount(), Refresh;
  HBRUSH old = (HBRUSH)GetCurrentObject(g_memDC, OBJ_BRUSH);

	// ***** calculate VU meter value and draw it as debounced value
  int last_a, last_b;
	int total_a = 0, total_b = 0;

  if (cfg.VuUsesWaveData)
  { // ***** use waveform data
    last_a  = this_mod->waveformData[0][0];
    last_b  = this_mod->waveformData[1][0];
	  for (x = 1; x < 576; x ++)
	  {
		  total_a += abs(last_a - (this_mod->waveformData[0][x]));
		  last_a = (this_mod->waveformData[0][x]);
		  total_b += abs(last_b - (this_mod->waveformData[1][x]));
		  last_b = (this_mod->waveformData[1][x]);
	  }
	  total_a /= (576/2); if (total_a > 255) total_a = 255;
	  total_b /= (576/2); if (total_b > 255) total_b = 255;
  }
  else
  {	// ***** use spectrum data
    for (x = 0; x < 576/2; x ++)
	  {
		  total_a += this_mod->spectrumData[0][x];
		  total_b += this_mod->spectrumData[1][x];
	  }
	  total_a /= (576/6); 
    if (total_a > 255) total_a = 255;
    total_b /= (576/6); 
    if (total_b > 255) total_b = 255;
  }

  if (this_mod->nCh > 1)
  { // ***** two stereo channels
    total = (total_a + total_b) / 2;
  }
  else
  { // ****** one mono channel
    total = total_a;
    total_b = total_a;
  }
  
  // ***** calculate debounced VU meter value
  lights.PeakHistory[0][lights.PeakIdx] = total_a;
  lights.PeakHistory[1][lights.PeakIdx] = total_b;
  if ( ++lights.PeakIdx >= cfg.DebounceCnt )
  {
    lights.PeakIdx   = 0;
    lights.PeakValid = true;
  }

  int curr_peak_a, curr_peak_b;

  if ( lights.PeakValid )
  {
    curr_peak_a = 0;
    curr_peak_b = 0;
    for (x = 0; x < cfg.DebounceCnt; x ++)
    {
      curr_peak_a += lights.PeakHistory[0][x];
      curr_peak_b += lights.PeakHistory[1][x];
    }
    curr_peak_a /= cfg.DebounceCnt;
    curr_peak_b /= cfg.DebounceCnt;
    curr_peak = (curr_peak_a + curr_peak_b) / 2;
	  // ***** clear VU meter background and draw both channels
	  Rectangle(g_memDC,0,VU_A_Y_POS, VU_MOD_X_SIZE, VU_INFO_Y_POS);
    SelectObject(g_memDC, g_red_brush);
	  Rectangle(g_memDC,0,VU_A_Y_POS, curr_peak_a,VU_B_Y_POS);
    SelectObject(g_memDC, g_green_brush);
	  Rectangle(g_memDC,0,VU_B_Y_POS, curr_peak_b, VU_INFO_Y_POS);

    // ***** test VU peak for update of lights
    if ( curr_peak < lights.LastCmpPeak[0] )
    {
      PeakDiff = lights.LastCmpPeak[0] - curr_peak;
    }
    else
    {
      PeakDiff = curr_peak - lights.LastCmpPeak[0];
    }
    if ( (PeakDiff >= cfg.CmpDiffVu) )
    {
      if ( (TickCnt-lights.LastTickCmp) >= cfg.RefreshTime/2 )
      {
        UpdateLigths = true;
      }
    }
    if ( !UpdateLigths )
    { // ***** test for update if timout over
      if ( (TickCnt-lights.LastTickCmp) >= cfg.RefreshTime )
      {
        if ( PeakDiff >= cfg.CmpDiffVu/2 )
        {
          UpdateLigths = true;
        }
        else
        { // ***** next update faster if no new value set
          lights.LastTickCmp = TickCnt - cfg.RefreshTime/2;
        }
      }
    }
  }

  // ***** time for new update
  if ( UpdateLigths )
  {
    BYTE Cnt, Mask, Next, Cmp;

    lights.LastTickCmp = TickCnt;
    // ***** clear background
    SelectObject(g_memDC, old);
	  Rectangle(g_memDC,0,0,VU_MOD_X_SIZE,VU_LAMP_HEIGHT);
    // ***** create new lights and take over value on good validation
    LoopTimeout = 0;
    do
    {
      Next = lights.Lights;
      Next ^= curr_peak;
      Next ^= rand()/13;              // create new value
      Cmp = lights.Lights ^ Next;
      Cnt = 0;
      Mask = 1;
      // ***** count nr of changed bits
      for (x = 0; x < 8; x ++)
      {
        if ( Cmp & Mask ) Cnt++;
        Mask *=2;
      }
      if ( Cnt > cfg.ChangeCnt )
      {
        Cnt = 0;
      }
      else
      {
        Cnt = 0;
        Mask = 1;
        // ***** count nr of active bits in next value
        for (x = 0; x < 8; x ++)
        {
          if ( Next & Mask ) Cnt++;
          Mask *=2;
        }
        if ( Cnt > cfg.MaxCnt )
        {
          Cnt = 0;
        }
        if ( Cnt < cfg.MinCnt )
        {
          Cnt = 0;
        }
      }
      // ***** avoid long looping with side effect for music thread
      if ( ++LoopTimeout > 100)
      {
        if (!Cnt)
        {
          Cnt = Next;
        }
      }
    }while( !Cnt );
    lights.Lights = Next;
    // ***** send to COM port
    if ( lights.hCommPort != INVALID_HANDLE_VALUE )
    {
#pragma warning(disable:4996)
		sprintf(lights.TxBuf, ">3L%2.2LX\n", lights.Lights);
#pragma warning(default:4996)
	  WriteComPort(lights.hCommPort, lights.TxBuf, 6); 
    }

    // ***** draw new value on screen
    SelectObject(g_memDC, g_yellow_brush);
    if ( Next & 128 ) Ellipse(g_memDC, 0*VU_LAMP_SIZE, VU_LAMP_Y_POS, (0+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 64  ) Ellipse(g_memDC, 1*VU_LAMP_SIZE, VU_LAMP_Y_POS, (1+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 32  ) Ellipse(g_memDC, 2*VU_LAMP_SIZE, VU_LAMP_Y_POS, (2+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 16  ) Ellipse(g_memDC, 3*VU_LAMP_SIZE, VU_LAMP_Y_POS, (3+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 8   ) Ellipse(g_memDC, 4*VU_LAMP_SIZE, VU_LAMP_Y_POS, (4+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 4   ) Ellipse(g_memDC, 5*VU_LAMP_SIZE, VU_LAMP_Y_POS, (5+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 2   ) Ellipse(g_memDC, 6*VU_LAMP_SIZE, VU_LAMP_Y_POS, (6+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( Next & 1   ) Ellipse(g_memDC, 7*VU_LAMP_SIZE, VU_LAMP_Y_POS, (7+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
  }


  Refresh = TickCnt - lights.LastTickCnt;
  // ***** append some space for erase backgrond of older (longer) text
  if ( lights.hCommPort != INVALID_HANDLE_VALUE )
  {
#pragma warning(disable:4996) 
    len = _stprintf(Info,"Refresh: %d ms    ", Refresh);
#pragma warning(default:4996) 
  }
  else
  {
#pragma warning(disable:4996) 
    len = _stprintf(Info,"no COM port !    ");
#pragma warning(default:4996) 
  }
  SelectObject(g_memDC, old);
  TextOut(g_memDC, 0, VU_INFO_Y_POS, Info, len);
  lights.LastTickCnt = TickCnt;

  { // copy doublebuffer to window
		HDC hdc = GetDC(g_hMainWnd);
		BitBlt(hdc, 0, 0, VU_MOD_X_SIZE, VU_MOD_Y_SIZE, g_memDC, 0, 0, SRCCOPY);
		ReleaseDC(g_hMainWnd,hdc);
	}
	return 0;
}


// ***************************************************************************
// cleanup (opposite of init()). 
// Destroys the window, unregisters the window class
void quit(struct winampVisModule *this_mod)
{
  SendMessage(this_mod->hwndParent, WM_WA_IPC, 0, IPC_SETVISWND);

	config_write(this_mod, true);		// write configuration (window position)
	SelectObject(g_memDC,g_oldBM);	// delete our doublebuffer
	DeleteObject(g_memDC);
	DeleteObject(g_memBM);	

	DeleteObject(g_red_brush);
	DeleteObject(g_green_brush);	
	DeleteObject(g_yellow_brush);

  CloseComPort(lights.hCommPort);

  // delete our window
  if (g_myWindowState.me) 
  {
    SetForegroundWindow(this_mod->hwndParent);
    DestroyWindow(g_myWindowState.me);
  }
	UnregisterClass(szAppName,this_mod->hDllInstance); // unregister window class
}


// ***************************************************************************
void next_preset()
{
}

// ***************************************************************************
void previous_preset()
{
}

// ***************************************************************************
void load_random_preset()
{
}

// ***************************************************************************
void set_random(bool flag)
{
  g_random_presets_flag = flag;
}

// ***************************************************************************
void go_fullscreen()
{
  if ( g_app_plugin )
  {
    if (SendMessage(g_app_plugin->hwndParent,WM_WA_IPC,0,IPC_IS_PLAYING_VIDEO)>1) 
    {
      MessageBox(g_app_plugin->hwndParent, "Can't go fullscreen while video is playing", "SVis", 0);
    }
    else
    {
      SendMessage(g_app_plugin->hwndParent,WM_WA_IPC,1,IPC_SET_VIS_FS_FLAG);
      // ... now do the work of actually going fullscreen ...
    }
  }
}

// ***************************************************************************
void go_windowed()
{
  if ( g_app_plugin )
  {
    SendMessage(g_app_plugin->hwndParent,WM_WA_IPC,0,IPC_SET_VIS_FS_FLAG);
  }
}

// ***************************************************************************
void open_configuration()
{
}

// ***************************************************************************
void open_popup_menu()
{
}

// ***************************************************************************
// window procedure for our application window
LRESULT CALLBACK WndAppProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_CREATE:		    
      if ( !cfg.CfgIsValid )
      { // ***** load settings from INI file
        config_read(g_app_plugin);
      }
      return 0;

		case WM_ERASEBKGND:   
      return 0;

		case WM_PAINT:
			{ // update from doublebuffer
				PAINTSTRUCT ps;
				RECT r;
				HDC hdc = BeginPaint(hwnd,&ps);
				GetClientRect(hwnd,&r);
				BitBlt(hdc,0,0,r.right,r.bottom,g_memDC,0,0,SRCCOPY);
        {
          RECT x={r.left+g_width, r.top, r.right, r.bottom};
          RECT y={r.left, r.top+g_height, r.right, r.bottom};
          FillRect(hdc, &x, (HBRUSH)GetStockObject(WHITE_BRUSH));
          FillRect(hdc, &y, (HBRUSH)GetStockObject(WHITE_BRUSH));
        }
				EndPaint(hwnd,&ps);
			}
		  return 0;

		case WM_DESTROY: 
      PostQuitMessage(0); 
      return 0;

    case WM_KEYDOWN: // pass keyboard messages to main winamp window (for processing)
		case WM_KEYUP:
			{	// get this_mod from our window's user data
				winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hwnd,GWL_USERDATA);
				PostMessage(this_mod->hwndParent,message,wParam,lParam);
			}
		  return 0;

		case WM_WINDOWPOSCHANGING:
			{	// get config_x and config_y for configuration
				RECT r;
				GetWindowRect(g_myWindowState.me,&r);
				cfg.config_x = r.left;
				cfg.config_y = r.top;
			}
		  return 0;

    case WM_COMMAND: 
      {
        int id = LOWORD(wParam);
        switch (id) 
        {
          // user clicked on 'next' preset button
          case ID_VIS_NEXT: 
            next_preset(); 
            break;

          // user clicked on 'previous' preset button
          case ID_VIS_PREV: 
            previous_preset(); 
            break;

          // user clicked on 'random' togglebutton
          case ID_VIS_RANDOM: 
            { // determine if we're switching random on or off or if Winamp is asking us about the state of our random flag
              bool flag = HIWORD(wParam) ? 1 : 0; 
              // are we being asked about the state of our random flag ?
              if (wParam >> 16 == 0xFFFF) 
              { // tell winamp about our state
                if ( g_app_plugin )
                {
                  SendMessage(g_app_plugin->hwndParent, WM_WA_IPC, g_random_presets_flag, IPC_CB_VISRANDOM);
                }
                break;
              }
              // changes random_preset_flag 
              set_random(flag); 
              // if we are turning random on, we should switch to a new random preset right away
              if (flag) load_random_preset();
              break;
            }
          case ID_VIS_FS: 
            go_fullscreen(); 
            break;
          case ID_VIS_CFG: 
            open_configuration(); 
            break;
          case ID_VIS_MENU: 
            open_popup_menu(); 
            break;
        } // switch (id)
        break;
      }
    default:
      break;
  } // switch (message)
	return DefWindowProc(hwnd,message,wParam,lParam);
}

// ***************************************************************************
// makes a INI-file in the winamp directory named "plugin.ini"
void config_getinifn(struct winampVisModule *this_mod, TCHAR *ini_file)
{	
/*
  char *INI_FILE = (char*)SendMessage(this_mod->hwndParent,WM_WA_IPC,0,IPC_GETINIDIRECTORY);
  strcpy(INI_FILE, ini_file);
	_tcscat(ini_file,"plugin.ini");
*/
	TCHAR *p;
	GetModuleFileName(this_mod->hDllInstance,ini_file,MAX_PATH);
	p=ini_file+strlen(ini_file);
	while (p >= ini_file && *p != '\\') p--;
	if (++p >= ini_file) *p = 0;
#pragma warning(disable:4996)
	_tcscat(ini_file,"plugin.ini");
#pragma warning(default:4996)
}


// ***************************************************************************
// load configuration settings
void config_read(struct winampVisModule *this_mod)
{
  TCHAR   Str[10], Key[20], Def[10];
  double  Val;
  int     i;
	TCHAR   ini_file[MAX_PATH];

	config_getinifn(this_mod,ini_file);

  if ( this_mod->hDllInstance )
  {
    cfg.CfgIsValid = true;
  }

  // screen X position and Y position, repsectively
  cfg.config_x = GetPrivateProfileInt(this_mod->description,"Screen_x", cfg.config_x,ini_file);
	cfg.config_y = GetPrivateProfileInt(this_mod->description,"Screen_y", cfg.config_y,ini_file);


  // no of com port
  cfg.ComPort = GetPrivateProfileInt(this_mod->description,"ComPort", cfg.ComPort,ini_file);
  // nr of dbounce loops
  cfg.DebounceCnt = GetPrivateProfileInt(this_mod->description,"DebounceCnt", cfg.DebounceCnt,ini_file);
  // time slice for value uptate
  cfg.RefreshTime = GetPrivateProfileInt(this_mod->description,"RefreshTime", cfg.RefreshTime,ini_file);
  // ***** spectrum alalyzer
  { // compare diff. for spectrum alalyzer
    cfg.CmpDiffSpectr = GetPrivateProfileInt(mod_spectr.description,"CmpDiffSpectr", cfg.CmpDiffSpectr,ini_file);
    // reducer multiplier for spectrum analyzer
    for (i=0; i<8; i++)
    {
#pragma warning(disable:4996) 
      _stprintf(Key, "ReduceSpec%d", i);
      _stprintf(Def, "%1.2lf", cfg.ReduceSpec[i]);
#pragma warning(default:4996) 
      GetPrivateProfileString(mod_spectr.description,Key, Def, Str, sizeof(Str)/sizeof(TCHAR), ini_file);
      // ***** use sscanf not in a dll !!!
      Val = _ttof((TCHAR*)&Str[0]);
      cfg.ReduceSpec[i] = Val;
    }

  }

  // ***** VU-meter
  { // compare diff. for VU-meter
    cfg.CmpDiffVu = GetPrivateProfileInt(mod_vu.description,"CmpDiffVu", cfg.CmpDiffVu, ini_file);
    // 0: use spectr. data, 1: waveform data
    i = GetPrivateProfileInt(mod_vu.description,"VuUsesWaveData", cfg.VuUsesWaveData, ini_file);
    if (i) { cfg.VuUsesWaveData = true; } else { cfg.VuUsesWaveData = false; }
    // nr of bits allowed for changing on update of light value
    cfg.ChangeCnt = GetPrivateProfileInt(mod_vu.description,"ChangeCnt", cfg.ChangeCnt, ini_file);
    // minimum of bits that must be ON
    cfg.MinCnt = GetPrivateProfileInt(mod_vu.description,"MinCnt", cfg.MinCnt,ini_file);
    // maximum of bits that can be ON
    cfg.MaxCnt = GetPrivateProfileInt(mod_vu.description,"MaxCnt", cfg.MaxCnt,ini_file);
  }
}


// ***************************************************************************
// store configuration settings
void config_write(struct winampVisModule *this_mod, bool bPos)
{
	TCHAR string[32], Key[20];
	TCHAR ini_file[MAX_PATH];
  int i;

	config_getinifn(this_mod,ini_file);

  if ( bPos )
  {
#pragma warning(disable:4996) 
	  _stprintf(string,"%d", cfg.config_x);
#pragma warning(default:4996) 
	  WritePrivateProfileString(this_mod->description,"Screen_x",string,ini_file);
#pragma warning(disable:4996) 
	  _stprintf(string,"%d", cfg.config_y);
#pragma warning(default:4996) 
	  WritePrivateProfileString(this_mod->description,"Screen_y",string,ini_file);
  }
  else
  {
    // no of com port
#pragma warning(disable:4996) 
	  _stprintf(string,"%d", cfg.ComPort);
#pragma warning(default:4996) 
    WritePrivateProfileString(this_mod->description,"ComPort", string, ini_file);
    // nr of dbounce loops
#pragma warning(disable:4996) 
	  _stprintf(string,"%d", cfg.DebounceCnt);
#pragma warning(default:4996) 
    WritePrivateProfileString(this_mod->description,"DebounceCnt", string, ini_file);
    // time slice for value uptate
#pragma warning(disable:4996) 
	  _stprintf(string,"%d", cfg.RefreshTime);
#pragma warning(default:4996) 
    WritePrivateProfileString(this_mod->description,"RefreshTime", string, ini_file);
    if ( this_mod == &mod_spectr )
    { // compare diff. for spectrum alalyzer
#pragma warning(disable:4996) 
	    _stprintf(string,"%d", cfg.CmpDiffSpectr);
#pragma warning(default:4996) 
      WritePrivateProfileString(this_mod->description,"CmpDiffSpectr", string, ini_file);
      // reducer multiplier for spectrum analyzer
      for (i=0; i<8; i++)
      {
#pragma warning(disable:4996) 
        _stprintf(Key, "ReduceSpec%d", i);
        _stprintf(string, "%1.2lf", cfg.ReduceSpec[i]);
#pragma warning(default:4996) 
        WritePrivateProfileString(this_mod->description, Key, string, ini_file);
      }
    }

    if ( this_mod == &mod_vu )
    { // compare diff. for VU-meter
#pragma warning(disable:4996) 
	    _stprintf(string,"%d", cfg.CmpDiffVu);
#pragma warning(default:4996) 
      WritePrivateProfileString(this_mod->description,"CmpDiffVu", string, ini_file);
      // 0: use spectr. data, 1: waveform data
#pragma warning(disable:4996) 
	    _stprintf(string,"%d", (int)cfg.VuUsesWaveData);
#pragma warning(default:4996) 
      WritePrivateProfileString(this_mod->description,"VuUsesWaveData", string, ini_file);
      // nr of bits allowed for changing on update of light value
#pragma warning(disable:4996) 
	    _stprintf(string,"%d", (int)cfg.ChangeCnt);
#pragma warning(default:4996) 
      WritePrivateProfileString(this_mod->description,"ChangeCnt", string, ini_file);
      // minimum of bits that must be ON
#pragma warning(disable:4996) 
	    _stprintf(string,"%d", (int)cfg.MinCnt);
#pragma warning(default:4996) 
      WritePrivateProfileString(this_mod->description,"MinCnt", string, ini_file);
      // maximum of bits that can be ON
#pragma warning(disable:4996) 
	    _stprintf(string,"%d", (int)cfg.MaxCnt);
#pragma warning(default:4996) 
      WritePrivateProfileString(this_mod->description,"MaxCnt", string, ini_file);
    }
  }
}
