// File:        Win32Serial.cpp 
// Copyright:   (c) 2009 by Frydaysoft.de
// Author:      SFr (frydaysoft@gmx.de)
// Proj:        light visualization plugin (winamp)
// Short Info:  Basic write functions of serial communication

#include <tchar.h>
#include <stdio.h>
//#include <afxwin.h>
#include "Win32Serial.h"



/****************************************************************************
  FUNCTION:  InitComPortConfig()

  PURPOSE:   Set default Com port settings. returns TRUE if sucessful else FALSE

*****************************************************************************/

BOOL InitComPortConfig(RS232_config &sio_cfg, int ComPort, int baudrate)
{
#ifdef AFX_MANAGE_STATE
  AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
  sio_cfg.dcbCommPort.DCBlength         = sizeof(DCB);
  sio_cfg.dcbCommPort.BaudRate          = CBR_9600;
  sio_cfg.dcbCommPort.fBinary           = FALSE;
  sio_cfg.dcbCommPort.fParity           = FALSE;
  sio_cfg.dcbCommPort.fOutxCtsFlow      = FALSE;
  sio_cfg.dcbCommPort.fOutxDsrFlow      = FALSE;
  sio_cfg.dcbCommPort.fDtrControl       = DTR_CONTROL_DISABLE;
  sio_cfg.dcbCommPort.fDsrSensitivity   = FALSE;
  sio_cfg.dcbCommPort.fTXContinueOnXoff = TRUE;
  sio_cfg.dcbCommPort.fOutX             = TRUE;
  sio_cfg.dcbCommPort.fInX              = TRUE;
  sio_cfg.dcbCommPort.fErrorChar        = FALSE;
  sio_cfg.dcbCommPort.fNull             = FALSE;
  sio_cfg.dcbCommPort.fRtsControl       = FALSE;
  sio_cfg.dcbCommPort.fAbortOnError     = FALSE;
  sio_cfg.dcbCommPort.XonLim            = MAX_RCV_QUEUE / 4 ;
  sio_cfg.dcbCommPort.XoffLim           = sio_cfg.dcbCommPort.XonLim;
  sio_cfg.dcbCommPort.ByteSize          = 8;
  sio_cfg.dcbCommPort.Parity            = NOPARITY;
  sio_cfg.dcbCommPort.StopBits          = ONESTOPBIT;
  sio_cfg.dcbCommPort.XonChar           = 17;
  sio_cfg.dcbCommPort.XoffChar          = 19;

  sio_cfg.dcbCommPort.StopBits = ONESTOPBIT;     // 1Stop
  sio_cfg.dcbCommPort.Parity   = NOPARITY;       // no par.
  sio_cfg.dcbCommPort.BaudRate = baudrate;       // 57600Bit
  sio_cfg.dcbCommPort.ByteSize = 8;              // 8bit
  sio_cfg.ComPort              = ComPort;        // COM 1

  return TRUE;
}


/****************************************************************************
  FUNCTION:  GetRS232ConfigString()

  PURPOSE:   Returns COM port settings as ascii string. returns ascii string

*****************************************************************************/

TCHAR* GetComPortString(RS232_config sio_cfg)
{
#ifdef AFX_MANAGE_STATE
    AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
  static TCHAR text[300];

#pragma warning(disable:4996)   
    _stprintf(text,"Com%d, ", sio_cfg.ComPort);

    switch(sio_cfg.dcbCommPort.BaudRate)
    {
        case CBR_9600:     _stprintf(text,"%s9600 Baud, "  ,text); break;
        case CBR_19200:    _stprintf(text,"%s19200 Baud, " ,text); break;
        case CBR_38400:    _stprintf(text,"%s38400 Baud, " ,text); break;
        case CBR_56000:    _stprintf(text,"%s56000 Baud, " ,text); break;
        case CBR_57600:    _stprintf(text,"%s57600 Baud, " ,text); break;
        case CBR_115200:   _stprintf(text,"%s115200 Baud, ",text); break;
        default:           _stprintf(text,"%s? Baud, "     ,text); break;
    }

    switch(sio_cfg.dcbCommPort.ByteSize)
    {
        case 7:      _stprintf(text,"%s7 Bit, ",text); break;
        case 8:      _stprintf(text,"%s8 Bit, ",text); break;
        default:     _stprintf(text,"%s? Bit, ",text); break;
    }

    switch(sio_cfg.dcbCommPort.StopBits)
    {
        case ONESTOPBIT:      _stprintf(text,"%s1 Stop, ",text); break;
        case TWOSTOPBITS:     _stprintf(text,"%s2 Stop, ",text); break;
        default:              _stprintf(text,"%s? Stop, ",text); break;
    }

    switch(sio_cfg.dcbCommPort.Parity)
    {
        case EVENPARITY:      _stprintf(text,"%seven Parity",text); break;
        case NOPARITY:        _stprintf(text,"%sno Parity"  ,text); break;
        case ODDPARITY:       _stprintf(text,"%sodd Parity" ,text); break;
        default:              _stprintf(text,"%s? Parity"   ,text); break;
    }
#pragma warning(default:4996)   
    
    return text;
}


/****************************************************************************
  FUNCTION:  OpenComPort()

  PURPOSE:   Open a Com port. returns TRUE if successful else FALSE

*****************************************************************************/
TCHAR* OpenComPort(HANDLE & hCommPort, RS232_config sio_cfg)
{
  static TCHAR  str_out[100];           // error text
  COMMTIMEOUTS  ctmoCommPort;           // some COM settings
  TCHAR         szPort[7];              // for max: "COM999"

#ifdef AFX_MANAGE_STATE
  AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
#pragma warning(disable:4996)   
  _stprintf(str_out,"");
#pragma warning(default:4996)   

  if ( hCommPort )
  {
    hCommPort = INVALID_HANDLE_VALUE;
  }
    
  // Check to see if an invalid command was sent
  if ( (sio_cfg.ComPort >=MAX_NUM_COMPORTS) || (sio_cfg.ComPort<1)) 
  {
#pragma warning(disable:4996)   
    _stprintf(str_out,"Given COM Port out of range");
#pragma warning(default:4996)   
    return str_out;
  }

//  if ( sio_cfg.ComPort < COM_NR_OTHER_STRING )
  {
#pragma warning(disable:4996)   
    _stprintf(szPort,"COM%d",sio_cfg.ComPort);
#pragma warning(default:4996)   
  }
    
  hCommPort = CreateFile(szPort, /*GENERIC_READ |*/ GENERIC_WRITE, 0,
                  NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
  
  // comport could not be opened
  if(hCommPort == INVALID_HANDLE_VALUE)
  {  
#pragma warning(disable:4996)   
        _stprintf(str_out,"INVLID_HANDLE_VALUE");
#pragma warning(default:4996)   
        return str_out;
  }

  // Set the time out values for the comport
  ctmoCommPort.ReadIntervalTimeout         = MAXDWORD;
  ctmoCommPort.ReadTotalTimeoutMultiplier  = 0;
  ctmoCommPort.ReadTotalTimeoutConstant    = 0 ;
  ctmoCommPort.WriteTotalTimeoutMultiplier = 0;
  ctmoCommPort.WriteTotalTimeoutConstant   = 0;
  if(!SetCommTimeouts(hCommPort, &ctmoCommPort))
  {
#pragma warning(disable:4996)   
    _stprintf(str_out,"Could not set comm timeouts");
#pragma warning(default:4996)   
    return str_out;
  }

  if(!SetCommState(hCommPort, &sio_cfg.dcbCommPort) )
  {
#pragma warning(disable:4996)   
    _stprintf(str_out,"Could not set the comstate");
#pragma warning(default:4996)   
    return str_out;
  }
  
  return str_out;
}


/****************************************************************************
  FUNCTION:  CloseComPort()

  PURPOSE:   Close a Com port and returns TRUE if sucessful. Return FALSE 
             elswise returns TRUE if succesful

*****************************************************************************/
BOOL CloseComPort(HANDLE & hCommPort)
{
#ifdef AFX_MANAGE_STATE
  AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
  if (hCommPort==INVALID_HANDLE_VALUE) 
  {
    return FALSE;
  }

  ClearTxBuff(hCommPort);
  if( CloseHandle(hCommPort) )
  { 
    hCommPort=INVALID_HANDLE_VALUE; 
    return TRUE; 
  }
  else 
  {
    return FALSE;
  }
}



/****************************************************************************
  FUNCTION:  ClearTxBuff()

  PURPOSE:   Clear the Driver Tx Buffer 
*****************************************************************************/
BOOL ClearTxBuff(const HANDLE & hCommPort)
{
#ifdef AFX_MANAGE_STATE
  AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
  if (hCommPort==INVALID_HANDLE_VALUE) return 0;

  // Clear the transmit buffer for character that were received
  return PurgeComm(hCommPort, PURGE_TXCLEAR|PURGE_TXABORT);    
}


/****************************************************************************
  FUNCTION:  WriteComPort()

  PURPOSE:   Write a command to the input buffer. Return the number of character
             placed in the TX buffer.
*****************************************************************************/
DWORD WriteComPort(  const HANDLE  & hCommPort, char *pData, int Len)
{
#ifdef AFX_MANAGE_STATE
  AFX_MANAGE_STATE(AfxGetStaticModuleState());
#endif
  if(hCommPort == INVALID_HANDLE_VALUE)
  {
    return 0;
  }
  DWORD dwAxisCount;

  if( !WriteFile(hCommPort, pData, Len, &dwAxisCount, NULL)  ) 
  {  
    return 0;
  }

  return (dwAxisCount);
}



