// ***************************************************************************
// File:        PipeServer.c
// Copyright:   (c) 2010 by Frydaysoft.de
// Author:      SFr (frydaysoft@gmx.de)
// Proj:        light visualization plugin (winamp)
// Short Info:  named pipe server application
// ***************************************************************************


#define PIPE_PATH "\\\\.\\pipe\\vislight"


//*************************************************************************

#define  STRICT
#include <conio.h>
#include <windows.h>
#include "PipeServer.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

PIPE_HANDLE gl_Client;
#define IDM_ABOUTBOX 0x6000


// yellow
#define COLOR_LAMP          RGB(255,255,0) 
#define COLOR_TEXT          RGB(128,128,128)
#define BK_COLOR            RGB(0,0,0)

#define VU_MOD_X_SIZE   256
#define VU_LAMP_SIZE    (VU_MOD_X_SIZE/8)
#define VU_LAMP_HEIGHT  (VU_LAMP_SIZE+2)
#define VU_LAMP_Y_POS   20
#define OUTER_FRM       25

#define WND_X VU_MOD_X_SIZE+10
#define WND_Y (VU_LAMP_Y_POS+2*VU_LAMP_HEIGHT+OUTER_FRM)


//*************************************************************************
int APIENTRY WinMain (HINSTANCE hInstance,
                      HINSTANCE hPrevInstance,
                      LPSTR  lpCmdLine,
                      int    nCmdShow)
{

  MSG           msg;
  WNDCLASS      wc;
  OSVERSIONINFO osvi;
  HMENU         hmenu;
  CHAR          tmpBuf[200];
  int           Val;

  UNREFERENCED_PARAMETER( lpCmdLine );
  UNREFERENCED_PARAMETER( hPrevInstance );


  // ***** Detect platform and exit gracefully if not Windows NT.
  osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
  GetVersionEx (&osvi);
  if (osvi.dwPlatformId != VER_PLATFORM_WIN32_NT)
  {
     LoadString(hInstance, IDS_WRONGOS, gl_Client.lpBuffer, sizeof(gl_Client.lpBuffer));  
     MessageBox (NULL, gl_Client.lpBuffer, "Pipe Server LightCtrl", MB_OK | MB_ICONSTOP);
     return 0;
  }

  gl_Client.brushBk   = CreateSolidBrush(BK_COLOR);
  gl_Client.brushLamp = CreateSolidBrush(COLOR_LAMP);
  gl_Client.hpen      = CreatePen(PS_SOLID, 1, COLOR_TEXT);
  gl_Client.hInst     = hInstance;
  gl_Client.hPipe     = 0;
  gl_Client.live      = 0;
  gl_Client.Msg[0]    = 0;
  gl_Client.Lights_0  = 0;
  gl_Client.Lights_1  = 0;
  gl_Client.OutPort   = 0;

  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC)MainWndProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInstance;
  wc.hIcon         = LoadIcon (hInstance, "npserver");
  wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
  wc.hbrBackground = gl_Client.brushBk; 
  wc.lpszMenuName  = "PipeMenu";
  wc.lpszClassName = "PipeWClass";

  RegisterClass(&wc);

  LoadString(hInstance, IDS_WINDOWTITLE, tmpBuf, sizeof(tmpBuf));
  if (strlen(lpCmdLine) >= strlen("-p xxx"))
  {
    if ( (lpCmdLine[0]=='-') && (lpCmdLine[1]=='p') && (lpCmdLine[2]==' '))
    {
      sscanf(&lpCmdLine[3], "%X", &Val);
      gl_Client.OutPort = (WORD)Val;
      //MessageBox (NULL, lpCmdLine, "Pipe Server LightCtrl", MB_OK | MB_ICONSTOP);
    }
  }
  wsprintf (gl_Client.lpBuffer, tmpBuf, gl_Client.OutPort);

  gl_Client.hWnd = CreateWindow("PipeWClass",
                      gl_Client.lpBuffer,
                      WS_OVERLAPPEDWINDOW,
                      (GetSystemMetrics(SM_CXFULLSCREEN)-WND_X)/2,
                      (GetSystemMetrics(SM_CYFULLSCREEN)-WND_Y)/2,
                      WND_X,
                      WND_Y,
                      NULL,
                      NULL,
                      gl_Client.hInst,
                      NULL);

  hmenu = GetSystemMenu(gl_Client.hWnd, FALSE);
  if (hmenu)
  {
	  AppendMenu(hmenu,  MF_SEPARATOR, 0, 0);
	  AppendMenu(hmenu,  MF_BYPOSITION | MF_STRING, IDM_ABOUTBOX, "About");
  }
  ShowWindow(gl_Client.hWnd, nCmdShow);

  while (GetMessage (&msg, NULL, 0, 0))
  {
    DispatchMessage (&msg);
  }

  DeleteObject(gl_Client.brushBk);
  DeleteObject(gl_Client.brushLamp);
  DeleteObject(gl_Client.hpen);

  return (msg.wParam);
}


//*************************************************************************
LONG CALLBACK MainWndProc (HWND   hwnd, UINT   message, WPARAM wParam, LPARAM lParam)
{
  LONG        lpServerThreadID;
  PAINTSTRUCT paintStruct;
  HDC         hDC;

  switch (message)
  {
    case WM_PAINT:
      // ***** Draw area
      hDC = BeginPaint (hwnd, &paintStruct);
      DrawView(hDC, &paintStruct.rcPaint);
      EndPaint (hwnd, &paintStruct);
      return (0);

    case WM_SYSCOMMAND:
      if (wParam == IDM_ABOUTBOX)
      {
        MessageBox (NULL, "Simple Pipe Server V1.01 for the LightCtrl plugin.\nReading winamp data from the pipe:\n"PIPE_PATH"\n\nStartup parameter:\n-p PortNr\ne.g. 378, 278, 3BC for LPTx or 000 for EMU\n\n(c) 2011 www.frydaysoft.de", "Pipe Server LightCtrl", MB_OK | MB_ICONINFORMATION);
      }
      break;

    case WM_CREATE :
      // ***** Create the first instance of a server side of the pipe.
      CreateThread ((LPSECURITY_ATTRIBUTES)NULL,       // No security.
                    (DWORD)0,                          // Same stack size.
                    (LPTHREAD_START_ROUTINE)ThreadProc,// Thread procedure.
                    (LPVOID)&hwnd,                     // Parameter.
                    (DWORD)0,                          // Start immediatly.
                    (LPDWORD)&lpServerThreadID);       // Thread ID.
      return (0);

    case WM_DESTROY :
      PostQuitMessage (0);
      return (0);
  }
  return DefWindowProc (hwnd, message, wParam, lParam);
}



//*************************************************************************
VOID ThreadProc(HWND *hWnd)
{
    CHAR   inBuf[IN_BUF_SIZE];     // Input buffer for pipe.
    CHAR   Msg[MSG_SIZE];          // buffer for a single msg

    CHAR   errorBuf[LINE_LEN] = "";     // Used for error messages.
    DWORD  bytesRead;                   // Used in ReadFile().
    DWORD  retCode;                     // Used to trap return codes.
    DWORD  i,j;

    // ***** Create the local named pipe
    gl_Client.live     = FALSE;
    gl_Client.Lights_0 = 0;
    gl_Client.Lights_1 = 0;
    TellMsg(">???");                    // Forces a paint

    gl_Client.hPipe = CreateNamedPipe (PIPE_PATH,              // Pipe name.
                            PIPE_ACCESS_INBOUND,              // 1 way pipe.
                            PIPE_WAIT                         // Wait on messages.
                            | PIPE_READMODE_BYTE ,            // Specify message mode pipe.
                            1,                                // Maximum instance limit.
                            OUT_BUF_SIZE,                     // Buffer sizes.
                            IN_BUF_SIZE,
                            TIME_OUT,                         // Specify time out.
                            NULL);                            // no Security attributes.

    // ***** Check Errors.
    if (gl_Client.hPipe == INVALID_HANDLE_VALUE)
    {
      retCode = GetLastError();         // Report any error, it should always succeed.
      LoadString(gl_Client.hInst, IDS_ERRORCODE, gl_Client.lpBuffer, sizeof(gl_Client.lpBuffer));
      wsprintf (errorBuf, gl_Client.lpBuffer, retCode);
      LoadString(gl_Client.hInst, IDS_DEBUGTITLE, gl_Client.lpBuffer, sizeof(gl_Client.lpBuffer));
      MessageBox (*hWnd, errorBuf, gl_Client.lpBuffer, MB_ICONINFORMATION | MB_OK | MB_APPLMODAL);
    }
    else
    {
      /*
      retCode = 0;
      while(1)
      {
        _sleep(500);
        sprintf(&Msg[0], ">3L%02X",(BYTE)++retCode);
        TellMsg(Msg);
      }
      */

      while (!gl_Client.live)
      {
        // ***** Block until a client connects.
        gl_Client.live = ConnectNamedPipe(gl_Client.hPipe, NULL);
   
        do
        {
          // ***** Read the pipe.
          if (retCode = ReadFile(gl_Client.hPipe, inBuf, sizeof(inBuf), &bytesRead, 0))
          { // ***** ">2JC\n"  or ">3Lxx\n"
            // --> use simplest synchronisation. 
            //     minimum delay between data blocks should be about 30ms
            //     This situation should be fine enough to read a single
            //     data block with every ReadFile() call.
            for (i=0, j=0; i < bytesRead; i++)
            {
              if (j)
              { // copy data til end block
                if (inBuf[i]=='\n') 
                { // found end
                  Msg[j]='\0';
                  TellMsg(Msg);
                  break;
                }
                Msg[j++] = inBuf[i];
                if ( j >= MSG_SIZE )
                {
                  j=0;
                  break;
                }
              }
              else
              { // found block start
                if (inBuf[i]=='>') 
                {
                  Msg[j++]=inBuf[i];
                }
              }
            }
          }
        }while (retCode);

        TellMsg(">xxx");                        // Forces a paint
        gl_Client.live = !DisconnectNamedPipe(gl_Client.hPipe);  // Close pipe instance.
      } // while (!gl_Client.live)
    }

    gl_Client.live = FALSE;                // Turns spool gray.
    CloseHandle(gl_Client.hPipe);          // Close handles.
    ExitThread(0);                         // Clean up and die.
}


//*************************************************************************
VOID TellMsg(CHAR *buffer)
{
  int Val;
  strcpy(gl_Client.Msg, buffer);
  if ( (gl_Client.Msg[0] == '>') &&
      ( (gl_Client.Msg[2] == 'L') && 
        ( (gl_Client.Msg[1] == '5') || (gl_Client.Msg[1] == '3') )
      )
     )
  { // ***** ">3Lxx\n"
    sscanf(&gl_Client.Msg[3], "%X", &Val);
    gl_Client.Lights_1 = (BYTE)(Val>>8);
    gl_Client.Lights_0 = (BYTE)Val;
    if (gl_Client.OutPort)
    {
      _outp(gl_Client.OutPort, gl_Client.Lights_0);
    }
  }
  InvalidateRect(gl_Client.hWnd, NULL, TRUE);
}


//*************************************************************************
VOID DrawView(HDC hDC, RECT *pRect)
{
  int    Idx;

  //FillRect(hDC, pRect, gl_Client.brushBk);

  // ***** Write msg
  SetTextColor(hDC, COLOR_TEXT);
  SetBkColor(hDC, BK_COLOR);
  //SetBkMode(hDC, OPAQUE);
  TextOut(hDC, 0, 0, gl_Client.Msg, strlen(gl_Client.Msg));

  SelectObject(hDC, gl_Client.hpen);
  for (Idx = 0; Idx < 8; Idx ++)
  {
    if ( gl_Client.Lights_0 & (1<<(7-Idx)) ) 
    {
      SelectObject(hDC, gl_Client.brushLamp);
    }
    else
    {
      SelectObject(hDC, gl_Client.brushBk);
    }
    Ellipse(hDC, Idx*VU_LAMP_SIZE, VU_LAMP_Y_POS, (Idx+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
    if ( gl_Client.Lights_1 & (1<<(7-Idx)) ) 
    {
      SelectObject(hDC, gl_Client.brushLamp);
    }
    else
    {
      SelectObject(hDC, gl_Client.brushBk);
    }
    Ellipse(hDC, Idx*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE, (Idx+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+2*VU_LAMP_SIZE);
  }
}
