// File:        LightVis.cpp 
// Copyright:   (c) 2012 by Frydaysoft.de
// Author:      SFr (frydaysoft@gmx.de)
// Proj:        light visualization plugin (winamp)
// Short Info:  Basic application and utility functions of the plugin

// ***************************************************************************
// * Winamp visualization plugin
// * This plugin analyzes the audio data and generates light pattern.
// * These pattern will be displayed on screen and transmitted to the 
// * LightUnit over the selected COM port 
// ***************************************************************************

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <time.h>
#include "winamp\wa_ipc.h"
#include "sio\Win32Serial.h"
#include "vis.h"
#include "resource.h"
#include <stdarg.h>
#include <Winuser.h>


//#define MOD_TRACE
#ifdef MOD_TRACE 
  static TCHAR dbg_trace[50];
#endif


// ***************************************************************************
#ifndef UDM_GETPOS 
#define UDM_GETPOS (WM_USER+113)
#endif
#ifndef UDM_SETRANGE 
#define UDM_SETRANGE 0x0465
#endif

// ***************************************************************************
#define _dbg_out(Info) //SetDlgItemText(hwndDlg,IDC_STATUS, Info)
//#define _dbg_out(Info) OutputDebugString(Info)

// ***************************************************************************
// ***** convert given (unicode) string into a double value 
#ifndef _ttof
  #ifndef _UNICODE
    // ***** use function from C-lib
    #define _ttof atof
  #else
    // ***** not part of the C-lib (only: _ttoi/_ttol)
    double _ttof(TCHAR *p)
    {
      #error "implementation tbd!"
      // e.g: convert to ansi and use atof()
    }
  #endif
#endif

// ***************************************************************************
typedef enum
{
  COL_BACK_IDX   = 0, // background color
  COL_TEXT_IDX   = 1, // text color (status/info)
  COL_LAMP_H_IDX = 2, // color higher 8 lamps
  COL_LAMP_L_IDX = 3, // color lower 8 lamps
  COL_BAR_H_IDX  = 4, // color for odd/even (spec) bar 
  COL_BAR_L_IDX  = 5, // or left/rigth (vu) bar
  COL_MAX_IDX    = 5  // limit constant for loop
} TColIdx;

// ***************************************************************************
#define LAMP_LEN 8
#define LAMP_CNT 2

// ***************************************************************************
// some constants for CFG (dialog)
#define SPEC_DEB_DEF              3
#define VU_DEB_DEF                3
#define DEB_MIN                   1
#define DEB_MAX                   6
#define DEB_WND_TEXT              "debounce (1..6)"
#define SPEC_RETRIG_DEF           60
#define VU_RETRIG_DEF             60
#define RETRIG_MIN                60
#define RETRIG_MAX                3000
#define RETRIG_WND_TEXT           "retrigger (60..3000ms)"
#define SPEC_CMP_DIFF_DEF         25 
#define SPEC_CMP_DIFF_MIN         2
#define SPEC_CMP_DIFF_MAX         120
#define SPEC_CMP_DIFF_WND_TEXT    "cmp_diff (2..120)"
#define VU_CMP_DIFF_DEF           60
#define VU_CMP_DIFF_MIN           10
#define VU_CMP_DIFF_MAX           200
#define VU_CMP_DIFF_WND_TEXT      "cmp_diff (10..200)"

#define VU_CHG_DIFF_DEF           5   // nr of changes for update
#define VU_CHG_DIFF_MIN           0
#define VU_CHG_DIFF_MAX           20
#define VU_CHANGE_DEF             2   // nr of bits that can be changed from old to new lights value
#define VU_CHANGE_MIN             0
#define VU_CHANGE_MAX             4
#define VU_LIGHTS_OFF_DEF         2   // nr of bits that must be ON in new lights value
#define VU_LIGHTS_OFF_MIN         0
#define VU_LIGHTS_OFF_MAX         5
#define VU_LIGHTS_ON_DEF          5   // maximum nr of bits that can be ON in new lights value
#define VU_LIGHTS_ON_MIN          0
#define VU_LIGHTS_ON_MAX          8
// defaults see SpecDefaultReduce[]
#define REDUCE_MIN                0.01  // allowed min
#define REDUCE_MAX                2.00  // allowed max

// ***************************************************************************
typedef struct
{
  COLORREF        rgb_col;
  TCHAR           cb_text[6+1];
}TComboInfo; 

// ***************************************************************************
// some constant paint position for the VU meter
#define START_STOP_X_OFFS   3
#define START_STOP_Y_OFFS   4
#define START_STOP_X_SIZE   5
#define START_STOP_Y_SIZE   10

#define VU_MOD_X_SIZE       256
#define VU_LAMP_SIZE        (VU_MOD_X_SIZE/(LAMP_CNT*LAMP_LEN))
#define VU_LAMP_HEIGHT      (VU_LAMP_SIZE+1)
#define VU_HEIGHT           24
#define VU_INFO_HEIGHT      32
#define VU_MOD_Y_SIZE       (VU_LAMP_HEIGHT + 2*VU_HEIGHT + VU_INFO_HEIGHT)
#define VU_LAMP_Y_POS       (0)
#define VU_A_Y_POS          (VU_LAMP_HEIGHT)
#define VU_B_Y_POS          (VU_A_Y_POS + VU_HEIGHT)
#define VU_INFO_Y_POS       (VU_B_Y_POS + VU_HEIGHT)
#define VU_INFO_X_POS       (START_STOP_X_OFFS+START_STOP_X_SIZE+5)
// some constant paint position for the spectrum analyzer
#define SPEC_MOD_X_SIZE     256
#define SPEC_LAMP_SIZE      (SPEC_MOD_X_SIZE/(LAMP_CNT*LAMP_LEN))
#define SPEC_LAMP_HEIGHT    (SPEC_LAMP_SIZE+2)
#define SPEC_HEIGHT         128  //(288/2)
#define SPEC_INFO_HEIGHT    32
#define SPEC_MOD_Y_SIZE     (SPEC_LAMP_HEIGHT + SPEC_HEIGHT + SPEC_INFO_HEIGHT)
#define SPEC_LAMP_Y_POS     (0)
#define SPEC_Y_POS          (SPEC_LAMP_HEIGHT)
#define SPEC_INFO_Y_POS     (SPEC_LAMP_HEIGHT + SPEC_HEIGHT)
#define SPEC_INFO_X_POS     VU_INFO_X_POS


// ***************************************************************************
COLORREF ModColors[COL_MAX_IDX+1];
COLORREF ModColorsTmp[COL_MAX_IDX+1];


// ***************************************************************************
// GUI IDs to access these controls
int UID_VuMeter[]      = { IDC_CHECK_USE_WAVE, IDC_EDIT_CH_DIFF_CNT, IDC_EDIT_CH_DIFF, IDC_EDIT_MIN, IDC_EDIT_MAX};
int UID_SpecAnalyzer[LAMP_CNT*LAMP_LEN]  = { IDC_EDIT_RED16, IDC_EDIT_RED15, IDC_EDIT_RED14, IDC_EDIT_RED13,
                                             IDC_EDIT_RED12, IDC_EDIT_RED11, IDC_EDIT_RED10, IDC_EDIT_RED9,
                                             IDC_EDIT_RED8,  IDC_EDIT_RED7,  IDC_EDIT_RED6,  IDC_EDIT_RED5,
                                             IDC_EDIT_RED4,  IDC_EDIT_RED3,  IDC_EDIT_RED2,  IDC_EDIT_RED1};
int ColorComboEditUid[] = 
{
  IDC_COMBO_BK_COLOR, IDC_COMBO_TEXT_COLOR, 
  IDC_COMBO_LH_COLOR, IDC_COMBO_LL_COLOR,
  IDC_COMBO_BH_COLOR, IDC_COMBO_BL_COLOR
};

// ***************************************************************************
double SpecDefaultReduce[LAMP_CNT][LAMP_LEN] = 
      // 0     1     2     3     4     5     6     7
     {{0.08, 0.15, 0.25, 0.30, 0.40, 0.45, 0.55, 0.75},
      // 0     1     2     3     4     5     6     7
      {0.08, 0.15, 0.25, 0.30, 0.40, 0.45, 0.55, 0.75}};


// ***************************************************************************
// "utility" functions
int  ScanComPorts(void);
void config_read(struct winampVisModule *this_mod);                      // reads the configuration
void config_write(struct winampVisModule *this_mod, bool bPos);          // writes the configuration
void config_getinifn(struct winampVisModule *this_mod, TCHAR *ini_file); // makes the .ini file filename
winampVisModule *getModule(int which);                                   // returns a winampVisModule when requested. Used in hdr, below

// ***************************************************************************
// "member" functions
void config(struct winampVisModule *this_mod);                          // configuration dialog
int  init(struct winampVisModule *this_mod);                            // initialization for module
int  renderSpec(struct winampVisModule *this_mod);                      // rendering for module 1
int  renderVu(struct winampVisModule *this_mod);                        // rendering for module 2
void quit(struct winampVisModule *this_mod);                            // deinitialization for module
// our procedure for the application window
LRESULT CALLBACK WndAppProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
// our procedure for the configuration window
LRESULT CALLBACK WndConfigProc(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam);

// ***************************************************************************
// text array with valid COM ports (for this PC)
#define TXT_COM_LEN 7                                 // "COM 256"
TCHAR *RealComPorts[MAX_NUM_COMPORTS][TXT_COM_LEN+1];

// ***************************************************************************
TCHAR szAppName[] = "LightVisualization";                               // Our window class, etc

struct
{
  winampVisModule   *app_plugin; 
  winampVisModule   *cfg_plugin;
  HWND              hLastFocus;
  bool              random_presets_flag;
  POINT             aptTriangle[4];
  POINT             aptBreak[4];
  POINT             aptClear[2];

  int               ModRefresh;
  bool              ModBreaking;
  bool              ModBreakInv;
  bool              ModBreakUpd;
  DWORD             ModBreakTickCnt;
  RS232_config      SioCfg;
  embedWindowState  WindowState;
  int               width;
  int               height;
  HWND              hMainWnd;        // main window handle
  HDC               memDC;           // memory device context
  HBITMAP           memBM;           // memory bitmap (for memDC)
  HBITMAP           oldBM;           // old bitmap (from memDC)
  HPEN              null_pen;
  HPEN              txt_pen;
  HPEN              bk_pen;
  HBRUSH            txt_brush;
  HBRUSH            bk_brush;
  HBRUSH            lamp_brush[LAMP_CNT][LAMP_LEN];
  HBRUSH            bars_brush[LAMP_CNT][LAMP_LEN];
} g_mod =
{
  NULL, NULL,
  0, 0,
  {0,0,  0,0,  0,0, 0,0},
  {0,0,  0,0,  0,0, 0,0}, 
  {0,0,  0,0},
};

// ***************************************************************************
// configuration declarations
struct
{
  bool            CfgIsValid;                         // true if config loaded from INI-File
  int             config_x, config_y;                 // screen X position and Y position, repsectively
  int             ComPort;                            // no of com port
  int             VuDebounceCnt;                      // nr of dbounce loops
  DWORD           VuRefreshTime;                      // time slice for value uptate
  int             VuCmpDiff;                          // compare diff. for VU-meter
  bool            VuUsesWaveData;                     // 0: use spectr. data, 1: waveform data
  BYTE            VuChgDiffCnt;                       // nr of changes for update
  BYTE            VuChangeCnt;                        // nr of bits allowed for changing on update of light value
  BYTE            VuMinCnt;                           // minimum of bits that must be ON
  BYTE            VuMaxCnt;                           // maximum of bits that can be ON
  int             SpecDebounceCnt;                    // nr of dbounce loops
  DWORD           SpecRefreshTime;                    // time slice for value uptate
  int             SpecCmpDiff;                        // compare diff. for spectrum alalyzer
  double          SpecReduce[LAMP_CNT][LAMP_LEN];     // reducer multiplier for spectrum analyzer
}
cfg_change,                                           // temp. cfg in CFG dialog
cfg_copy,                                             // backup cfg in CFG dialog
cfg =                                                 // active cfg 
{ 
    false,                                              // CfgIsValid
    50, 50,                                             // config_x, config_y
    PIPE_IDX,                                           // ComPort
    VU_DEB_DEF,                                         // DebounceCnt
    VU_RETRIG_DEF,                                      // RetriggerSlot
    VU_CMP_DIFF_DEF,                                    // CmpDiffVu
    false,                                              // VuUsesWaveData
    VU_CHG_DIFF_DEF, VU_CHANGE_DEF,                     // VuChgDiffCnt, ChangeCnt
    VU_LIGHTS_OFF_DEF, VU_LIGHTS_ON_DEF,                // ChangeCnt, MinCnt, MaxCnt
    SPEC_DEB_DEF,                                       // DebounceCnt
    SPEC_RETRIG_DEF,                                    // RetriggerSlot
    SPEC_CMP_DIFF_DEF,                                  // CmpDiffSpectr
    {{SpecDefaultReduce[0][0], 
      SpecDefaultReduce[0][1], 
      SpecDefaultReduce[0][2], 
      SpecDefaultReduce[0][3], 
      SpecDefaultReduce[0][4], 
      SpecDefaultReduce[0][5], 
      SpecDefaultReduce[0][6], 
      SpecDefaultReduce[0][7]},
     {SpecDefaultReduce[1][0], 
      SpecDefaultReduce[1][1], 
      SpecDefaultReduce[1][2], 
      SpecDefaultReduce[1][3], 
      SpecDefaultReduce[1][4], 
      SpecDefaultReduce[1][5], 
      SpecDefaultReduce[1][6], 
      SpecDefaultReduce[1][7]}}                       // SpecReduce
};

// ***************************************************************************
TComboInfo cfg_colors[] =
{
  // ***** gray
  {RGB( 0xFF, 0xFF, 0xFF ), _T("xxxxxx")  },
  {RGB( 0xC0, 0xC0, 0xC0 ), _T("xxxxxx")  },
  {RGB( 0x80, 0x80, 0x80 ), _T("xxxxxx")  },
  {RGB( 0x40, 0x40, 0x40 ), _T("xxxxxx")  },
  {RGB( 0x00, 0x00, 0x00 ), _T("xxxxxx")  },
  // ***** blue
  {RGB( 0x00, 0x00, 0x40 ), _T("xxxxxx")  },
  {RGB( 0x00, 0x00, 0x80 ), _T("xxxxxx")  },
  {RGB( 0x00, 0x00, 0xC0 ), _T("xxxxxx")  },
  {RGB( 0x00, 0x00, 0xFF ), _T("xxxxxx")  },
  // green
  {RGB( 0x00, 0x40, 0x00 ), _T("xxxxxx")  },
  {RGB( 0x00, 0x80, 0x00 ), _T("xxxxxx")  },
  {RGB( 0x00, 0xC0, 0x00 ), _T("xxxxxx")  },
  {RGB( 0x00, 0xFF, 0x00 ), _T("xxxxxx")  },
  // red
  {RGB( 0x40, 0x00, 0x00 ), _T("xxxxxx")  },
  {RGB( 0x80, 0x00, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xC0, 0x00, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0x00, 0x00 ), _T("xxxxxx")  },

  {RGB( 0x40, 0x40, 0x00 ), _T("xxxxxx")  },
  {RGB( 0x80, 0x80, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xC0, 0xC0, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0xFF, 0x00 ), _T("xxxxxx")  },

  {RGB( 0x00, 0x40, 0x40 ), _T("xxxxxx")  },
  {RGB( 0x00, 0x80, 0x80 ), _T("xxxxxx")  },
  {RGB( 0x00, 0xC0, 0xC0 ), _T("xxxxxx")  },
  {RGB( 0x00, 0xFF, 0xFF ), _T("xxxxxx")  },

  {RGB( 0x40, 0x00, 0x40 ), _T("xxxxxx")  },
  {RGB( 0x80, 0x00, 0x80 ), _T("xxxxxx")  },
  {RGB( 0xC0, 0x00, 0xC0 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0x00, 0xFF ), _T("xxxxxx")  },

  {RGB( 0xFF, 0x40, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0x80, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0xC0, 0x00 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0xFF, 0x00 ), _T("xxxxxx")  },

  {RGB( 0xFF, 0x00, 0x40 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0x00, 0x80 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0x00, 0xC0 ), _T("xxxxxx")  },
  {RGB( 0xFF, 0x00, 0xFF ), _T("xxxxxx")  },
  
  {RGB( 0x00, 0x40, 0xFF ), _T("xxxxxx")  },
  {RGB( 0x00, 0x80, 0xFF ), _T("xxxxxx")  },
  {RGB( 0x00, 0xC0, 0xFF ), _T("xxxxxx")  },
  {RGB( 0x00, 0xFF, 0xFF ), _T("xxxxxx")  },

};

#define CFG_COLOR_CNT (sizeof(cfg_colors)/sizeof(TComboInfo))
HBRUSH  cfg_brush[CFG_COLOR_CNT];

// ***************************************************************************
#define TX_MAX_LEN  8     // ">3L??\n" or ">5L????\n" or ">2JC\n"
struct
{
  HANDLE          hCommPort;
  BYTE            ChgDiffCnt;
  DWORD           LastTickCnt;                              // for time calculation
  DWORD           LastTickCmp;                              // compare time to detect next time slice for update of light value
  int             LastCmpPeak[LAMP_CNT][LAMP_LEN];          // 2 for VU, 2x8 for spectrum
  BYTE            LightVal[LAMP_CNT];                       // current value for LightUnit
  char            TxBuf[TX_MAX_LEN+1];                      // tx buffer
  int             PeakHistory[LAMP_CNT][LAMP_LEN][DEB_MAX]; // 2 for VU, 2x8 for spectrum
  int             PeakIdx;                                  // index into PeakHistory[]
  bool            PeakValid;                                // true if PeakHistory[] filled
}lights = 
{
  INVALID_HANDLE_VALUE,                     // hCommPort
  0,                                        // ChgDiffCnt
  0,                                        // LastTickCnt
  0,                                        // LastTickCmp
  {{0},{0}},                                // LastCmpPeak
  {0,0},                                    // LightVal[2]
  {0},                                      // TxBuf[]
  {0},                                      // PeakHistory[]
  0,                                        // PeakIdx
  0,                                        // PeakValid
};


// ***************************************************************************
// Module header, includes version, description, and address of the module retriever function
winampVisHeader hdr = { VIS_HDRVER, "Light Visualisation v1.3 (www.frydaysoft.de)", getModule };


// ***************************************************************************
// module: spectrum analyzer
winampVisModule mod_spectr =
{
  "Light Spectrum Analyzer",  // Description
  NULL,                       // hwndParent:    Parent window (filled in by calling app)
  NULL,                       // hDllInstance:  DLL instance (filled in by calling app)
  0,                          // sRate:         Sample rate (filled in by calling app)
  0,                          // nCh:           Number of channels (filled in by calling app)
  25,                         // latencyMS:     Latency from call of RenderFrame to actual drawing
  25,                         // delayMS:       Delay between calls in ms
  2,                          // spectrumNch:   Number of Spectrum Analysis Channels
  0,                          // waveformNch:   Number of Waveform Analysis Channels
  { 0, },                     // spectrumData[2][576]
  { 0, },                     // waveformData[2][576]
  config,                     // Configuration dialog 
  init,                       // Initialization (0 on success, 1 if vis should end)
  renderSpec,                 // Render (0 on success, 1 if vis should end)
  quit                        // Exit procedure
};

// ***************************************************************************
// module: VU meter
winampVisModule mod_vu =
{
  "Light VU Meter",           // Description
  NULL,                       // hwndParent:    Parent window (filled in by calling app)
  NULL,                       // hDllInstance:  DLL instance (filled in by calling app)
  0,                          // sRate:         Sample rate (filled in by calling app)
  0,                          // nCh:           Number of channels (filled in by calling app)
  25,                         // latencyMS:     Latency from call of RenderFrame to actual drawing
  25,                         // delayMS:       Delay between calls in ms
  2,                          // spectrumNch:   Number of Spectrum Analysis Channels
  2,                          // waveformNch:   Number of Waveform Analysis Channels
  { 0, },                     // spectrumData[2][576]
  { 0, },                     // waveformData[2][576]
  config,                     // Configuration dialog 
  init,                       // Initialization (0 on success, 1 if vis should end)
  renderVu,                   // Render (0 on success, 1 if vis should end)
  quit                        // Exit procedure
};


//******************************************************************************
#if 1
  int sprintf_char(char *buf, const char *fmt, ...)
  {
    va_list args;
    int ret;

    va_start(args, fmt); // Initialize the va_list
    #pragma warning(disable:4996)
    ret = vsprintf(buf, fmt, args);
    #pragma warning(default:4996)
    va_end(args);       // Cleanup the va_list
    return ret;
  }

  #ifdef  _UNICODE
    int stprintf(wchar_t *buf, const wchar_t *fmt, ...)
    {
      va_list args;
      int ret;
      va_start(args, fmt); // Initialize the va_list
      #pragma warning(disable:4996)
      ret = vswprintf(buf, fmt, args);
      #pragma warning(default:4996)
      va_end(args);       // Cleanup the va_list
      return ret;
    }
  #else
    #define stprintf  sprintf
  #endif
#else
  #define sprintf_char    sprintf
  #define stprintf        _stprintf
#endif


// ***************************************************************************
// this is the only exported symbol. returns our main header.
// if you are compiling C++, the extern "C" { is necessary, so we just #ifdef it
#ifdef __cplusplus
extern "C" {
#endif
__declspec( dllexport ) winampVisHeader *winampVisGetHeader()
{
  return &hdr;
}
#ifdef __cplusplus
}
#endif

// ***************************************************************************
#ifdef __cplusplus
extern "C" {
#endif
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  static int singleton = 0;
  // fdwReason
  // 1: start/init
  // 2: processing
  // 3: stop/end

  if (!singleton)
  {
#ifdef MOD_TRACE 
    dbg_trace[0] = '\0';
#endif
    g_mod.ModRefresh  = 0;
    g_mod.ModBreaking = 0;
    g_mod.ModBreakInv = false;
    g_mod.ModBreakUpd = false;
    g_mod.ModBreakTickCnt = GetTickCount();
    for (int i=0; i<CFG_COLOR_CNT; i++)
    {
      cfg_brush[i] = CreateSolidBrush(cfg_colors[i].rgb_col);
      stprintf(cfg_colors[i].cb_text, "%6.6lX", cfg_colors[i].rgb_col);
    }
    singleton = 1;
  }

  if (fdwReason == 1)
  {
    lights.LastTickCmp = lights.LastTickCnt = GetTickCount();
  }

  return TRUE;
}
#ifdef __cplusplus
}
#endif



// ***************************************************************************
// getmodule routine from the main header. Returns NULL if an invalid module was requested,
// otherwise returns either mod_spectr or mod_vu depending on 'which'.
winampVisModule *getModule(int which)
{
  switch (which)
  {
    case 0: return &mod_spectr;
    case 1: return &mod_vu;
    default:return NULL;
  }
}

// ***************************************************************************
typedef enum
{
  COL_INIT     = 0,
  COL_UPDATE   = 1,
  COL_DELETE   = 2,
  COL_STORE    = 3,
  COL_RESTORE  = 4,
  COL_WR_2_INI = 5
}TAction;

int process_colors(struct winampVisModule *this_mod, TAction action)
{
  struct
  {
    BITMAPINFOHEADER bmiHeader;
    ULONG bmiColors[2];
    DWORD aj[8];
  } mybmi;
  int   y,x;
  TCHAR string[32], Key[20];
  TCHAR ini_file[MAX_PATH];
  DWORD col;

  switch ( action )
  {
    case COL_INIT:
      if (this_mod == &mod_vu)
      {
        ModColors[COL_BACK_IDX]   = 0x000000;
        ModColors[COL_TEXT_IDX]   = RGB(0x80,0x80,0x80);
        ModColors[COL_LAMP_H_IDX] = RGB(0xC0,0xC0,0x00);
        ModColors[COL_LAMP_L_IDX] = RGB(0xFF,0xFF,0x00);
        ModColors[COL_BAR_H_IDX]  = RGB(0x00,0x80,0x00);
        ModColors[COL_BAR_L_IDX]  = RGB(0x00,0xFF,0x00);
        stprintf(Key, "Vu");
      }
      else
      {
        ModColors[COL_BACK_IDX]   = 0x000000;
        ModColors[COL_TEXT_IDX]   = RGB(0x80,0x80,0x80);
        ModColors[COL_LAMP_H_IDX] = RGB(0xC0,0xC0,0x00);
        ModColors[COL_LAMP_L_IDX] = RGB(0xFF,0xFF,0x00);
        ModColors[COL_BAR_H_IDX]  = RGB(0x00,0x80,0x00);
        ModColors[COL_BAR_L_IDX]  = RGB(0x00,0xFF,0x00);
        stprintf(Key, "Spec");
      }
      config_getinifn(this_mod,ini_file);
      for (x=0; x<=COL_MAX_IDX; x++)
      {
        stprintf(string, "%s_ColorIdx_%d", Key, x);
        //ModColors[x] = GetPrivateProfileInt(this_mod->description, string, ModColors[x], ini_file);
      }
      break;
    case COL_UPDATE:
      g_mod.ModRefresh = 2;
    case COL_DELETE:
      break;
    case COL_STORE:
      for (x=0; x<=COL_MAX_IDX; x++)
      {
        ModColorsTmp[x] = ModColors[x];
      }
      return 0;
    case COL_RESTORE:
      for (x=0; x<=COL_MAX_IDX; x++)
      {
        ModColors[x] = ModColorsTmp[x];
      }
      return 0;
    case COL_WR_2_INI:
      config_getinifn(this_mod,ini_file);
      if (this_mod == &mod_vu)
      {
        for (x=0; x<=COL_MAX_IDX; x++)
        {
          stprintf(Key, "Vu_ColorIdx_%d", x);
          stprintf(string,"%d", ModColors[x]);
          WritePrivateProfileString(this_mod->description,Key,string,ini_file);
        }
      }
      else
      {
        for (x=0; x<=COL_MAX_IDX; x++)
        {
          stprintf(Key, "Spec_ColorIdx_%d", x);
          stprintf(string,"%d", ModColors[x]);
          WritePrivateProfileString(this_mod->description,Key,string,ini_file);
        }
      }
      return 0;
    default: 
      return 1;
      break;
  }

  if ( (action == COL_UPDATE) || (action == COL_DELETE) )
  {
    DeleteObject(g_mod.null_pen);
    DeleteObject(g_mod.txt_pen);
    DeleteObject(g_mod.bk_pen);
    DeleteObject(g_mod.txt_brush);
    DeleteObject(g_mod.bk_brush);
    for (y=0; y<LAMP_CNT; y++)
    {
      for (x=0; x<LAMP_LEN; x++)
      {
        DeleteObject(g_mod.bars_brush[y][x]); 
        DeleteObject(g_mod.lamp_brush[y][x]);
      }
    }
  }

  if ( action < COL_DELETE )
  {  
    g_mod.null_pen       = CreatePen(PS_NULL,1,RGB(0,0x80,0));
    g_mod.txt_pen        = CreatePen(PS_SOLID,2,ModColors[COL_TEXT_IDX]);
    g_mod.bk_pen         = CreatePen(PS_SOLID,2,ModColors[COL_BACK_IDX]);
    g_mod.txt_brush      = CreateSolidBrush( ModColors[COL_TEXT_IDX] );
    g_mod.bk_brush       = CreateSolidBrush( ModColors[COL_BACK_IDX] );
    
    // ***** some static brushes for faster painting
    ZeroMemory(&mybmi, sizeof mybmi);
    mybmi.bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
    mybmi.bmiHeader.biWidth         = 8;
    mybmi.bmiHeader.biHeight        = -8;
    mybmi.bmiHeader.biPlanes        = 1;
    mybmi.bmiHeader.biBitCount      = 1;
    mybmi.bmiHeader.biCompression   = BI_RGB;
    mybmi.bmiHeader.biSizeImage     = 0;
    mybmi.bmiHeader.biXPelsPerMeter = 0;
    mybmi.bmiHeader.biYPelsPerMeter = 0;
    mybmi.bmiHeader.biClrUsed       = 0;
    mybmi.bmiHeader.biClrImportant  = 0;

    if (this_mod == &mod_vu)
    {
      mybmi.aj[0x00] = 0x0000FF;
      mybmi.aj[0x01] = 0xCC00CC;
      mybmi.aj[0x02] = 0xCC00CC;
      mybmi.aj[0x03] = 0xCC00CC;
      mybmi.aj[0x04] = 0xCC00CC;
      mybmi.aj[0x05] = 0xCC00CC;
      mybmi.aj[0x06] = 0xCC00CC;
      mybmi.aj[0x07] = 0xCC00CC;
    }
    else
    {
      mybmi.aj[0x00] = 0xFFFF;
      mybmi.aj[0x01] = 0x8080;
      mybmi.aj[0x02] = 0x8080;
      mybmi.aj[0x03] = 0xFFFF;
      mybmi.aj[0x04] = 0xFFFF;
      mybmi.aj[0x05] = 0x8080;
      mybmi.aj[0x06] = 0x8080;
      mybmi.aj[0x07] = 0xFFFF;
    }

    col = ModColors[COL_BACK_IDX];
    mybmi.bmiColors[1] = RGB( GetBValue(col),
                              GetGValue(col),
                              GetRValue(col) );
    for (y=0; y<(LAMP_CNT); y++)
    {
      for (x=0; x<(LAMP_LEN); x++)
      {
        col = ModColors[COL_BAR_H_IDX+(x & 1)];
        mybmi.bmiColors[0] = RGB( GetBValue(col),
                                  GetGValue(col),
                                  GetRValue(col) );
        g_mod.bars_brush[y][x] = CreateDIBPatternBrushPt(&mybmi, DIB_RGB_COLORS);
        g_mod.lamp_brush[y][x] = CreateSolidBrush( ModColors[COL_LAMP_H_IDX+y] );
      }
    }

  }
  return 0;
}

//******************************************************************************
// scans physical COM ports, existing ports will be noticed into string array
int ScanComPorts(void)
{
  TCHAR ComString[7];              // for max: "COM999"
  int   cnt = 0;

  memset(RealComPorts, 0, sizeof(RealComPorts));

  stprintf((TCHAR*)&RealComPorts[cnt], "COM %d", cnt);
  cnt++;

  for (int i=1; i<MAX_NUM_COMPORTS; i++)
  {
    stprintf(ComString, "COM%d", i);

    BOOL bSuccess = FALSE;
    HANDLE hPort = ::CreateFile(ComString, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
    if (hPort == INVALID_HANDLE_VALUE)
    {
      DWORD dwError = GetLastError();
      if (dwError == ERROR_ACCESS_DENIED || dwError == ERROR_GEN_FAILURE || dwError == ERROR_SHARING_VIOLATION || dwError == ERROR_SEM_TIMEOUT)
        bSuccess = TRUE;
    }
    else
    {
      bSuccess = TRUE;
      CloseHandle(hPort);
    }

    if (bSuccess)
    {
      stprintf((TCHAR*)&RealComPorts[cnt], "COM %d", i);
      cnt++;
    }
  }

  return cnt;
}


// ***************************************************************************
void _TryOnlineUpdate(HWND hwndDlg, int CtrlID)
{
  static TCHAR Info[50];
  int i;
  double val;
  if (IsDlgButtonChecked(hwndDlg,IDC_CHECK_ONLINE) == BST_CHECKED)
  {
    GetDlgItemText(hwndDlg, IDC_EDIT_DEB, Info, sizeof(Info)/sizeof(TCHAR));
    i = _ttoi(Info);
    if ( (i >= DEB_MIN) && (i <= DEB_MAX) )
    { 
      if (g_mod.cfg_plugin == &mod_vu)
      {
        cfg.VuDebounceCnt = i; 
      }
      else
      {
        cfg.SpecDebounceCnt = i; 
      }
    }
    GetDlgItemText(hwndDlg, IDC_EDIT_RETRIG, Info, sizeof(Info)/sizeof(TCHAR));
    i = _ttoi(Info);
    if ( (i >= RETRIG_MIN) && (i <= RETRIG_MAX) )
    { 
      if (g_mod.cfg_plugin == &mod_vu)
      {
        cfg.VuRefreshTime = i; 
      }
      else
      {
        cfg.SpecRefreshTime = i; 
      }
    }
    GetDlgItemText(hwndDlg, IDC_EDIT_CMP_DIFF, Info, sizeof(Info)/sizeof(TCHAR));
    i = _ttoi(Info);
    if (g_mod.cfg_plugin == &mod_vu)
    {
      if ( (i >= VU_CMP_DIFF_MIN) && (i <= VU_CMP_DIFF_MAX) )
      { 
        cfg.VuCmpDiff = i; 
      }
    }
    else
    {
      if ( (i >= SPEC_CMP_DIFF_MIN) && (i <= SPEC_CMP_DIFF_MAX) )
      {
        cfg.SpecCmpDiff = i; 
      }
    }
  
    //bool VuUsesWaveData;  // 0: use spectr. data, 1: waveform data

    if (g_mod.cfg_plugin == &mod_vu)
    {
      GetDlgItemText(hwndDlg, IDC_EDIT_CH_DIFF_CNT, Info, sizeof(Info)/sizeof(TCHAR));
      i = _ttoi(Info);
      if ( (i >= VU_CHG_DIFF_MIN) && (i <= VU_CHG_DIFF_MAX) )
      {
        cfg.VuChgDiffCnt = i; 
      }      
      GetDlgItemText(hwndDlg, IDC_EDIT_CH_DIFF, Info, sizeof(Info)/sizeof(TCHAR));
      i = _ttoi(Info);
      if ( (i >= VU_CHANGE_MIN) && (i <= VU_CHANGE_MAX) )
      {
        cfg.VuChangeCnt = i; 
      }
      GetDlgItemText(hwndDlg, IDC_EDIT_MIN, Info, sizeof(Info)/sizeof(TCHAR));
      i = _ttoi(Info);
      if ( (i >= VU_LIGHTS_OFF_MIN) && (i <= VU_LIGHTS_OFF_MAX) )
      {
        cfg.VuMinCnt = i; 
      }
      GetDlgItemText(hwndDlg, IDC_EDIT_MAX, Info, sizeof(Info)/sizeof(TCHAR));
      i = _ttoi(Info);
      if ( (i >= VU_LIGHTS_ON_MIN) && (i <= VU_LIGHTS_ON_MAX) )
      {
        cfg.VuMaxCnt = i; 
      }
    }
    if (g_mod.cfg_plugin == &mod_spectr)
    {
      // ***** spectrum analyzer
      for (i = 0; i < sizeof(UID_SpecAnalyzer)/sizeof(int); i++)
      {
        GetDlgItemText(hwndDlg, UID_SpecAnalyzer[i], Info, sizeof(Info)/sizeof(TCHAR));
        // ***** use sscanf not in a dll !!!
        val = _ttof(Info);
        if ( (val > REDUCE_MAX) || (val < REDUCE_MIN) )
        { 
          continue; 
        }
        if (!CtrlID)
        {
          cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] = val;  
        }
        else
        {
          if (CtrlID == UID_SpecAnalyzer[i])
          {
            cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] = val;
            stprintf(Info, "Reduce[%d] set to %1.2lf", i, cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN]);
            SetDlgItemText(hwndDlg,IDC_STATUS, Info);
            return;
          }
        }
      }
    }
  }
}

// ***************************************************************************
DWORD DecodeColRef(TCHAR *Info)
{
  TCHAR ch;
  DWORD ret = 0;
  if (Info)
  {
    for (int i = 0; i <= 7; i++)
    {
      ch = Info[i];
      if (!ch)
      {
        return ret;
      }
      else
      {
        ret <<= 4;
        if ( (ch <= '9') && (ch >= '0') )
        {
          ret += (ch-'0');
        }
        else
        {
          if ( (ch <= 'f') && (ch >= 'a') )
          {
            ret += (10+ch-'a');
          }
          else
          {
            if ( (ch <= 'F') && (ch >= 'A') )
            {
              ret += (10+ch-'A');
            }
            else
            {
              break;
            }
          }
        }
      }
    }
  }
  return 0xFF000000;
}


// ***************************************************************************
LRESULT CALLBACK WndConfigProc(HWND hwndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  static TCHAR Info[50], ComStr[10];
  static POINT pt={1,1};
  TCHAR  *pBuf;
  HWND   ComboHwnd, Tmp;
  int    i, PortSel, Val, Min, Max;
  DWORD  dw;
  int    LastFocusID;
  double edit_val;

  //stprintf(Info, "msg: %x, w: %X, l: %X", uMsg, wParam, lParam);
  //_dbg_out(Info);

  switch (uMsg)
  {
    // ***************************************************************************
    case WM_INITDIALOG:
    {
        if ( !cfg.CfgIsValid )
        { // ***** load settings from INI file
          config_read(g_mod.cfg_plugin);
        }
        cfg_change = cfg;

        SetWindowText(hwndDlg, "config dialog");

        // ***** COM port
        ComboHwnd = GetDlgItem(hwndDlg, IDC_COMBO_SEL);
        if ( ComboHwnd )
        {
          PortSel = 0;
          if ( lights.hCommPort != INVALID_HANDLE_VALUE )
          {
            stprintf(ComStr, "Com%d", g_mod.SioCfg.ComPort);
            SendMessage(ComboHwnd, CB_ADDSTRING, (WPARAM)0, (LPARAM)&ComStr[0]);
            ::EnableWindow(ComboHwnd, false);
            PortSel = 0;
          }
          else
          {
            for (i = 0; i < MAX_NUM_COMPORTS; i++)
            {
              pBuf = (TCHAR*)&RealComPorts[i][0];
              if ( *pBuf )
              {
                // ***** use sscanf not in a dll !!!
                Val = _ttoi(&pBuf[sizeof("COM")]);
                if ( Val == cfg.ComPort )
                {
                  PortSel = i;
                }
                SendMessage(ComboHwnd, CB_ADDSTRING, (WPARAM)0, (LPARAM)&RealComPorts[i]);
              }
              else
              {
                break;
              }
            }
          }
        }
        SendMessage(ComboHwnd, CB_SETCURSEL, (WPARAM)PortSel, (LPARAM)0);

        // ***** common configuration

        SetDlgItemText(hwndDlg,IDC_TEXT_DEB,      DEB_WND_TEXT);
        SetDlgItemText(hwndDlg,IDC_TEXT_RETRIG,   RETRIG_WND_TEXT);
        if (g_mod.cfg_plugin == &mod_spectr)
        {
          SetDlgItemInt(hwndDlg,IDC_EDIT_DEB,       cfg.SpecDebounceCnt, FALSE);
          SetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG,    cfg.SpecRefreshTime, FALSE);

          SetDlgItemText(hwndDlg,IDC_TEXT_CMP_DIFF, SPEC_CMP_DIFF_WND_TEXT);
          SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  cfg.SpecCmpDiff, FALSE);
          for (i = 0; i < sizeof(UID_VuMeter)/sizeof(int); i++)
          {
            Tmp = GetDlgItem(hwndDlg, UID_VuMeter[i]);
            ::EnableWindow(Tmp, FALSE);
          }
        }
        if (g_mod.cfg_plugin == &mod_vu)
        {
          SetDlgItemInt(hwndDlg,IDC_EDIT_DEB,       cfg.VuDebounceCnt, FALSE);
          SetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG,    cfg.VuRefreshTime, FALSE);

          SetDlgItemText(hwndDlg,IDC_TEXT_CMP_DIFF, VU_CMP_DIFF_WND_TEXT);
          SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  cfg.VuCmpDiff, FALSE);
          for (i = 0; i < sizeof(UID_SpecAnalyzer)/sizeof(int); i++)
          {
            Tmp = GetDlgItem(hwndDlg, UID_SpecAnalyzer[i]);
            ::EnableWindow(Tmp, FALSE);
          }
          Tmp = GetDlgItem(hwndDlg, IDC_SPIN_ANA);
          ::EnableWindow(Tmp, FALSE);
          Tmp = GetDlgItem(hwndDlg, IDC_STATIC_SPEC_GRP);
          if (Tmp) ::EnableWindow(Tmp, FALSE);
          // ***** status line
          SetDlgItemText(hwndDlg,IDC_STATUS, "configure: VU-meter");

        }
        else
        {
          Tmp = GetDlgItem(hwndDlg, IDC_SPIN_VU);
          ::EnableWindow(Tmp, FALSE);
          Tmp = GetDlgItem(hwndDlg, IDC_STATIC_VU_GRP);
          if (Tmp) ::EnableWindow(Tmp, FALSE);
          // ***** status line
          SetDlgItemText(hwndDlg,IDC_STATUS, "configure: Spectrum Analyzer");
        }
        // ***** spectrum analyzer
        for (i = 0; i < sizeof(UID_SpecAnalyzer)/sizeof(int); i++)
        {
          stprintf(Info, "%1.2lf", cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN]);
          SetDlgItemText(hwndDlg, UID_SpecAnalyzer[i], Info);
        }

        // ***** VU-meter
        CheckDlgButton(hwndDlg,IDC_CHECK_USE_WAVE, cfg.VuUsesWaveData?BST_CHECKED:BST_UNCHECKED);
        SetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF_CNT,  cfg.VuChgDiffCnt, FALSE);
        SetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF,      cfg.VuChangeCnt, FALSE);
        SetDlgItemInt(hwndDlg,IDC_EDIT_MIN,          cfg.VuMinCnt, FALSE);
        SetDlgItemInt(hwndDlg,IDC_EDIT_MAX,          cfg.VuMaxCnt, FALSE);


        SendDlgItemMessage(hwndDlg, IDC_SPIN_COMMON, UDM_SETRANGE, 0,MAKELPARAM(1, 0));
        SendDlgItemMessage(hwndDlg, IDC_SPIN_ANA, UDM_SETRANGE, 0,MAKELPARAM(1, 0));
        SendDlgItemMessage(hwndDlg, IDC_SPIN_VU, UDM_SETRANGE, 0,MAKELPARAM(1, 0));
        CheckDlgButton(hwndDlg, IDC_CHECK_ONLINE, BST_CHECKED);
        g_mod.hLastFocus = (HWND)0xDEFDEF;
 
        // ***** fill colors into combo boxes
        for (i = 0; i < sizeof(ColorComboEditUid)/sizeof(int); i++)
        {
          ComboHwnd = GetDlgItem(hwndDlg, ColorComboEditUid[i]);
          if ( ComboHwnd )
          {
            for (PortSel = 0; PortSel < CFG_COLOR_CNT; PortSel++)
            {           
              Val = SendMessage(ComboHwnd, CB_ADDSTRING, (WPARAM)0, (LPARAM)&cfg_colors[PortSel].cb_text[0]);
            }
          }
          stprintf(Info, "%6.6X", ModColors[i]);
          Val = SendMessage(ComboHwnd, CB_FINDSTRING, (WPARAM)-1,  (LPARAM)Info);
          if ( Val != CB_ERR )
          {
            SendMessage(ComboHwnd, CB_SETCURSEL, (WPARAM)Val,  0);
          }
          else
          {
            SendMessage(ComboHwnd, WM_SETTEXT, (WPARAM)0,  (LPARAM)Info);
          }
        }
      }
    break;
    case WM_CTLCOLOREDIT:
      for (i = 0; i < sizeof(ColorComboEditUid)/sizeof(int); i++)
      {
        ComboHwnd = GetDlgItem(hwndDlg, ColorComboEditUid[i]);
        Tmp = GetWindow((HDC)ComboHwnd, GW_CHILD);
        if ( (HDC)lParam == Tmp )
        {
            i = SendMessage(ComboHwnd, CB_GETCURSEL, 0, 0);
            if ( i != -1) 
            {
              SetBkMode((HDC)wParam, BKMODE_LAST);
              SetTextColor((HDC)wParam, (~cfg_colors[i].rgb_col)&0xFFFFFF);
			        SetBkColor((HDC)wParam, cfg_colors[i].rgb_col);
              return (LRESULT)cfg_brush[i];
            }
            break;
        }
      }
      // ***** mark edit field when editable by spin control (value up/down)
      if ( g_mod.hLastFocus && (g_mod.hLastFocus == (HWND)lParam) )
      {
			  SetTextColor((HDC)wParam, RGB(0, 0, 0));
			  SetBkColor((HDC)wParam, RGB(255, 255, 255));
        return (LRESULT)GetStockObject(WHITE_BRUSH);
      }
      // no break here !!!
    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORBTN:
    case WM_CTLCOLORDLG:
			//SetBkMode((HDC)wParam, BKMODE_LAST);
			SetTextColor((HDC)wParam, RGB(0, 255, 0));
      SetBkColor((HDC)wParam, RGB(0, 0, 0));
      return (LRESULT)( GetStockObject( BLACK_BRUSH ) ) ;
      break;
    //case WM_CTLCOLORLISTBOX:
    //  break;
    case WM_DRAWITEM:
      for (i = 0; i < sizeof(ColorComboEditUid)/sizeof(int); i++)
      {
        if ( (int)wParam == ColorComboEditUid[i] )
        {
          i = ((LPDRAWITEMSTRUCT)(lParam))->itemID;
          FillRect( ((LPDRAWITEMSTRUCT)(lParam))->hDC , &((LPDRAWITEMSTRUCT)(lParam))->rcItem , cfg_brush[i]);
          break;
        }
      }
      break;

    // ***************************************************************************
    case WM_SETCURSOR:
    {
      if ( g_mod.hLastFocus == (HWND)0xDEFDEF )
      {
        g_mod.hLastFocus = GetFocus();
      }
      if (LOWORD(lParam) == HTCLIENT)               // hit-test code
      {
        if ( (HIWORD(lParam) == WM_LBUTTONDOWN) )   // identifier of the mouse message
        {
          LastFocusID = GetDlgCtrlID((HWND)wParam);
          switch(LastFocusID)
          {
            case IDC_SPIN_COMMON:
            case IDC_SPIN_ANA:
            case IDC_SPIN_VU:
              break;
            default:
              if (g_mod.hLastFocus) 
              {
                //RedrawWindow(g_mod.hLastFocus, NULL, NULL, RDW_ERASE | RDW_FRAME | RDW_INVALIDATE | RDW_ALLCHILDREN);
                RedrawWindow(g_mod.hLastFocus, NULL, NULL, 0 | RDW_INVALIDATE );
              }
              g_mod.hLastFocus = (HWND)wParam; 
              break;
          }
        }
      }
      break;
    }
    // ***************************************************************************
    case WM_VSCROLL:
    {
            char buf[40];

            if (LOWORD(wParam) == SB_THUMBPOSITION)
            {
              int pos;

              switch(GetDlgCtrlID((HWND)lParam))
              {
                case IDC_SPIN_COMMON:
                  if (g_mod.hLastFocus)
                  {
                    pos=LOWORD(SendDlgItemMessage(hwndDlg, IDC_SPIN_COMMON, UDM_GETPOS, 0,0));
                    LastFocusID = GetDlgCtrlID(g_mod.hLastFocus);
                    switch (LastFocusID)
                    {
                      case IDC_EDIT_DEB:
                        Min = DEB_MIN;              Max = DEB_MAX;
                        break;
                      case IDC_EDIT_RETRIG:
                        Min = RETRIG_MIN;           Max = RETRIG_MAX;
                        break;
                      case IDC_EDIT_CMP_DIFF:
                        if (g_mod.cfg_plugin == &mod_vu)
                        {
                          Min = VU_CMP_DIFF_MIN;    Max = VU_CMP_DIFF_MAX;
                        }
                        else
                        {
                          Min = SPEC_CMP_DIFF_MIN;  Max = SPEC_CMP_DIFF_MAX;
                        }
                        break;
                      default:
                        Max = 0;
                        break;
                    }
                    if (Max)
                    {
                      GetDlgItemText(hwndDlg, LastFocusID, Info, sizeof(Info)/sizeof(TCHAR));
                      Val = _ttoi(Info) + (2*(pos&1)-1);
                      if (Val < Min )
                      { // ***** allowed min
                        Val = Min;
                      }
                      if (Val > Max )
                      { // ***** allowed max
                        Val = Max;
                      }
                      sprintf_char(buf, "%d", Val);
                      SetDlgItemText(hwndDlg,LastFocusID, buf);
                      SetFocus(g_mod.hLastFocus);
                      _TryOnlineUpdate(hwndDlg, LastFocusID);

                    }
                  }
                  break;
                case IDC_SPIN_ANA:
                  if (g_mod.hLastFocus)
                  {
                    pos=LOWORD(SendDlgItemMessage(hwndDlg, IDC_SPIN_ANA, UDM_GETPOS, 0,0));
                    LastFocusID = GetDlgCtrlID(g_mod.hLastFocus);
                    if ( (LastFocusID <= IDC_EDIT_RED16) && ((LastFocusID >= IDC_EDIT_RED1)) )
                    {
                      GetDlgItemText(hwndDlg, LastFocusID, Info, sizeof(Info)/sizeof(TCHAR));
                      edit_val = _ttof(Info) + (0.02*(double)(pos&1)-0.01);
                      if (edit_val < REDUCE_MIN )
                      { // ***** allowed min
                        edit_val = REDUCE_MIN;
                      }
                      if (edit_val > REDUCE_MAX )
                      { // ***** allowed max
                        edit_val = REDUCE_MAX;
                      }
                      sprintf_char(buf, "%0.2f", edit_val);
                      SetDlgItemText(hwndDlg,LastFocusID, buf);
                      SetFocus(g_mod.hLastFocus);
                      _TryOnlineUpdate(hwndDlg, LastFocusID);
                    }
                  }
                  break;
                case IDC_SPIN_VU:
                  if (g_mod.hLastFocus)
                  {
                    pos=LOWORD(SendDlgItemMessage(hwndDlg, IDC_SPIN_VU, UDM_GETPOS, 0,0));
                    LastFocusID = GetDlgCtrlID(g_mod.hLastFocus);
                    switch(LastFocusID)
                    {
                      case IDC_EDIT_CH_DIFF_CNT:
                        Min = VU_CHG_DIFF_MIN;
                        Max = VU_CHG_DIFF_MAX;
                        break;
                      case IDC_EDIT_CH_DIFF:
                        Min = VU_CHANGE_MIN;
                        Max = VU_CHANGE_MAX;
                        break;
                      case IDC_EDIT_MIN:
                        Min = VU_LIGHTS_OFF_MIN;
                        Max = VU_LIGHTS_OFF_MAX;
                        break;
                      case IDC_EDIT_MAX:
                        Min = VU_LIGHTS_ON_MIN;
                        Max = VU_LIGHTS_ON_MAX;
                        break;
                      default:
                        LastFocusID = 0;
                        break;
                    }
                    if (LastFocusID)
                    {
                      GetDlgItemText(hwndDlg, LastFocusID, Info, sizeof(Info)/sizeof(TCHAR));
                      Val = _ttoi(Info) + (2*(pos&1)-1);
                      if (Val < Min )
                      { // ***** allowed min
                        Val = Min;
                      }
                      if (Val > Max )
                      { // ***** allowed max
                        Val = Max;
                      }
                      sprintf_char(buf, "%d", Val);
                      SetDlgItemText(hwndDlg,LastFocusID, buf);
                      SetFocus(g_mod.hLastFocus);
                      _TryOnlineUpdate(hwndDlg, LastFocusID);
                    }
                  }
                  break;

                default:
                  break;
              }
            }

      }
    break;
    // ***************************************************************************
    case WM_COMMAND:
    {
      switch (LOWORD(wParam))
      {
        case IDC_BTN_LD_DEF:
          // ***** relaod defaults
          // ***** spectrum analyzer
          if (g_mod.cfg_plugin == &mod_spectr)
          {
            SetDlgItemInt(hwndDlg,IDC_EDIT_DEB,       SPEC_DEB_DEF, FALSE);
            SetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG,    SPEC_RETRIG_DEF, FALSE);
            SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  SPEC_CMP_DIFF_DEF, FALSE);
            cfg_change.SpecCmpDiff = SPEC_CMP_DIFF_DEF;
            for (i = 0; i < sizeof(UID_SpecAnalyzer)/sizeof(int); i++)
            {
              cfg_change.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] = SpecDefaultReduce[i/LAMP_LEN][i%LAMP_LEN];
              stprintf((TCHAR*)&Info, "%1.2lf", SpecDefaultReduce[i/LAMP_LEN][i%LAMP_LEN]);
              SetDlgItemText(hwndDlg, UID_SpecAnalyzer[i], Info);
            }
          }
          // ***** VU-meter
          if (g_mod.cfg_plugin == &mod_vu)
          {
            SetDlgItemInt(hwndDlg,IDC_EDIT_DEB,       VU_DEB_DEF, FALSE);
            SetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG,    VU_RETRIG_DEF, FALSE);
            SetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF,  VU_CMP_DIFF_DEF, FALSE);
            cfg_change.VuCmpDiff = VU_CMP_DIFF_DEF;
            CheckDlgButton(hwndDlg,IDC_CHECK_USE_WAVE, BST_UNCHECKED);
            cfg_change.VuUsesWaveData = false;
            SetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF_CNT,  VU_CHG_DIFF_DEF, FALSE);
            cfg_change.VuChangeCnt = VU_CHG_DIFF_DEF;
            SetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF,  VU_CHANGE_DEF, FALSE);
            cfg_change.VuChangeCnt = VU_CHANGE_DEF;
            SetDlgItemInt(hwndDlg,IDC_EDIT_MIN,  VU_LIGHTS_OFF_DEF, FALSE);
            cfg_change.VuMinCnt = VU_LIGHTS_OFF_DEF;
            SetDlgItemInt(hwndDlg,IDC_EDIT_MAX,  VU_LIGHTS_ON_DEF, FALSE);
            cfg_change.VuMaxCnt = VU_LIGHTS_ON_DEF;
          }
          _TryOnlineUpdate(hwndDlg, 0);
          break;
        case IDCANCEL:        // 2
          cfg = cfg_copy;
          process_colors(g_mod.cfg_plugin, COL_RESTORE);
          EndDialog(hwndDlg,0);
          break;
        case IDOK:            // 1
          ComboHwnd = GetDlgItem(hwndDlg, IDC_COMBO_SEL);
          if ( ComboHwnd )
          {
            // ***** COM port
            if ( IsWindowEnabled(ComboHwnd) )
            {
              i = SendMessage(ComboHwnd, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
              stprintf((TCHAR*)&Info, "sel: %d", i);
              i = SendMessage(ComboHwnd, CB_GETLBTEXT, (WPARAM)i, (LPARAM)Info);
              if ( i != CB_ERR )
              {
                // ***** use sscanf not in a dll !!!
                Val = _ttoi((TCHAR*)&Info[sizeof("COM")]);
                cfg_change.ComPort = Val;                
              }
            }

            if (g_mod.cfg_plugin == &mod_spectr)
            {
              cfg_change.SpecDebounceCnt = GetDlgItemInt(hwndDlg,IDC_EDIT_DEB, NULL, FALSE);
              if ( (cfg_change.SpecDebounceCnt > DEB_MAX) || (cfg_change.SpecDebounceCnt < DEB_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"debounce out of range"); break; }
              cfg_change.SpecRefreshTime = GetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG, NULL, FALSE);
              if ( (cfg_change.SpecRefreshTime > RETRIG_MAX) || (cfg_change.SpecRefreshTime < RETRIG_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"retrigger out of range"); break; }

              
              cfg_change.SpecCmpDiff = GetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF, NULL, FALSE);
              if ( (cfg_change.SpecCmpDiff > SPEC_CMP_DIFF_MAX) || (cfg_change.SpecCmpDiff < SPEC_CMP_DIFF_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"cmp_diff out of range"); break; }
            }
            if (g_mod.cfg_plugin == &mod_vu)
            {
              cfg_change.VuDebounceCnt = GetDlgItemInt(hwndDlg,IDC_EDIT_DEB, NULL, FALSE);
              if ( (cfg_change.VuDebounceCnt > DEB_MAX) || (cfg_change.VuDebounceCnt < DEB_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"debounce out of range"); break; }
              cfg_change.VuRefreshTime = GetDlgItemInt(hwndDlg,IDC_EDIT_RETRIG, NULL, FALSE);
              if ( (cfg_change.VuRefreshTime > RETRIG_MAX) || (cfg_change.VuRefreshTime < RETRIG_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"retrigger out of range"); break; }


              cfg_change.VuCmpDiff = GetDlgItemInt(hwndDlg,IDC_EDIT_CMP_DIFF, NULL, FALSE);
              if ( (cfg_change.VuCmpDiff > VU_CMP_DIFF_MAX) || (cfg_change.VuCmpDiff < VU_CMP_DIFF_MIN) )
              { SetDlgItemText(hwndDlg,IDC_STATUS,"cmp_diff out of range"); break; }
            }

            // ***** spectrum analyzer
            for (i = 0; i < sizeof(UID_SpecAnalyzer)/sizeof(int); i++)
            {
              GetDlgItemText(hwndDlg, UID_SpecAnalyzer[i], Info, sizeof(Info)/sizeof(TCHAR));
              // ***** use sscanf not in a dll !!!
              cfg_change.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] = _ttof(Info);
              if ( (cfg_change.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] > REDUCE_MAX) || (cfg_change.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] < REDUCE_MIN) )
              { 
                stprintf(Info,"reduce beam(%d, %1.2f) out of range(%1.2f..%1.2f)",8-i,cfg_change.SpecReduce[i/LAMP_LEN][i%LAMP_LEN],REDUCE_MIN,REDUCE_MAX); 
                SetDlgItemText(hwndDlg,IDC_STATUS,Info); 
                break; 
              }
            }
            if (i != (sizeof(UID_SpecAnalyzer)/sizeof(int)) )
            {
              break;
            }

            // ***** VU-meter
            if ( IsDlgButtonChecked(hwndDlg,IDC_CHECK_USE_WAVE) == BST_CHECKED )
            { cfg_change.VuUsesWaveData = true; } else { cfg_change.VuUsesWaveData = false; }
            cfg_change.VuChgDiffCnt = GetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF_CNT,  NULL, FALSE);
            if ( (cfg_change.VuChgDiffCnt > VU_CHG_DIFF_MAX) || (cfg_change.VuChgDiffCnt < VU_CHG_DIFF_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"change_cnt of range"); break; }
            cfg_change.VuChangeCnt = GetDlgItemInt(hwndDlg,IDC_EDIT_CH_DIFF,  NULL, FALSE);
            if ( (cfg_change.VuChangeCnt > VU_CHANGE_MAX) || (cfg_change.VuChangeCnt < VU_CHANGE_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"change_diff of range"); break; }
            cfg_change.VuMinCnt    = GetDlgItemInt(hwndDlg,IDC_EDIT_MIN,  NULL, FALSE);
            if ( (cfg_change.VuMinCnt > VU_LIGHTS_OFF_MAX) || (cfg_change.VuMinCnt < VU_LIGHTS_OFF_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"min_lights of range"); break; }
            cfg_change.VuMaxCnt    = GetDlgItemInt(hwndDlg,IDC_EDIT_MAX,  NULL, FALSE);
            if ( (cfg_change.VuMaxCnt > VU_LIGHTS_ON_MAX) || (cfg_change.VuMaxCnt < VU_LIGHTS_ON_MIN) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"max_lights of range"); break; }
            if ( (cfg_change.VuMinCnt > cfg_change.VuMaxCnt ) )
            { SetDlgItemText(hwndDlg,IDC_STATUS,"min_lights > max_lights"); break; }
          }
          // ***** store settings
          cfg = cfg_change;
          process_colors(g_mod.cfg_plugin, COL_WR_2_INI);
          config_write(g_mod.cfg_plugin, false); // write cfg part
          EndDialog(hwndDlg,0);
          break;
        default:
          if (IsDlgButtonChecked(hwndDlg,IDC_CHECK_ONLINE) == BST_CHECKED)
          {
            for (i = 0; i < sizeof(ColorComboEditUid)/sizeof(int); i++)
            {
              if (LOWORD(wParam) == ColorComboEditUid[i])
              {
                ComboHwnd = GetDlgItem(hwndDlg, ColorComboEditUid[i]);
                if ( ComboHwnd )
                {
                  switch(HIWORD(wParam))
                  {
                    case CBN_SELENDOK:
                      Val = SendMessage(ComboHwnd, CB_GETCURSEL, 0, 0);
                      if ( Val != -1) 
                      {
                        ModColors[i] = cfg_colors[Val].rgb_col;
                        process_colors(g_mod.cfg_plugin, COL_UPDATE);
                      }
                      break;
                    case CBN_EDITUPDATE:
                      if ( SendMessage(ComboHwnd, WM_GETTEXT, (WPARAM)7,  (LPARAM)Info) )
                      {
                        dw = DecodeColRef(Info);
                        if ( dw <= 0xFFFFFF )
                        {
                          ModColors[i] = dw;
                          process_colors(g_mod.cfg_plugin, COL_UPDATE);
                        }
                      }

                    default:
                      break;
                  }
                }
                break;
              }
            }
          }
          break;
      }
      break;
    }
    default:
      break;
  }
  return 0;
}


// ***************************************************************************
void config(struct winampVisModule *this_mod)
{
  g_mod.cfg_plugin = this_mod;
  ScanComPorts();
  process_colors(this_mod, COL_INIT);
  cfg_copy = cfg;
  process_colors(g_mod.cfg_plugin, COL_STORE);
  DialogBox(g_mod.cfg_plugin->hDllInstance,MAKEINTRESOURCE(IDD_DIALOG1),g_mod.cfg_plugin->hwndParent, (DLGPROC)WndConfigProc);
  process_colors(g_mod.cfg_plugin, COL_UPDATE);
}



// ***************************************************************************
// initialization. Registers our window class, creates our window, etc. 
// Note: this one works for both modules, but you should make init1() and init2()...
// returns 0 on success, 1 on failure.
int init(struct winampVisModule *this_mod)
{
  int styles;
  HWND parent = NULL;
  HWND (*e)(embedWindowState *v);
  BITMAPINFO bmi;
  void *bits = NULL;

  // ***** set some global utility data members
  g_mod.app_plugin = this_mod;
  g_mod.width = (this_mod == &mod_vu)?VU_MOD_X_SIZE:SPEC_MOD_X_SIZE;
  g_mod.height = (this_mod == &mod_vu)?(VU_MOD_Y_SIZE):SPEC_MOD_Y_SIZE;

  // ***** basic data initialization
  lights.LastTickCmp = GetTickCount();
  lights.LastTickCnt = GetTickCount();
  lights.ChgDiffCnt = 0;
  memset(&lights.LastCmpPeak, 0, sizeof(lights.LastCmpPeak));
  srand( (unsigned)time( NULL ) );
  lights.LightVal[0] = 0;
  lights.LightVal[1] = 0;
  lights.PeakIdx     = 0;
  lights.PeakValid   = 0;
  lights.hCommPort   = INVALID_HANDLE_VALUE;


  config_read(this_mod);

  g_mod.WindowState.flags = 0;
  // uncomment this line if your plugin draws to the screen using directx OVERLAY mode
  // g_mod.WindowState.flags |= EMBED_FLAGS_NOTRANSPARENCY; 
  g_mod.WindowState.flags |= EMBED_FLAGS_NORESIZE;

  //cfg.config_x = 10;
  //cfg.config_y = 10;

  g_mod.WindowState.r.left    = cfg.config_x;
  g_mod.WindowState.r.top     = cfg.config_y;

  g_mod.WindowState.r.right   = cfg.config_x + g_mod.width;
  if (this_mod == &mod_spectr)
  {
    // ***** note: winamp resizes the window in tabbed steps so set the height 
    //             a little bit greather for the spectrum analyzer
    g_mod.WindowState.r.bottom  = cfg.config_y + g_mod.height + 50;
    g_mod.aptTriangle[0].y = SPEC_INFO_Y_POS;    
  }
  else
  {
    g_mod.WindowState.r.bottom  = cfg.config_y + g_mod.height;
    g_mod.aptTriangle[0].y = VU_INFO_Y_POS;    
  }

  g_mod.aptTriangle[0].x = START_STOP_X_OFFS;
  g_mod.aptTriangle[0].y += START_STOP_Y_OFFS;
  g_mod.aptTriangle[1].x = g_mod.aptTriangle[0].x + START_STOP_X_SIZE;
  g_mod.aptTriangle[1].y = g_mod.aptTriangle[0].y + (START_STOP_Y_SIZE/2);
  g_mod.aptTriangle[2].x = g_mod.aptTriangle[0].x;
  g_mod.aptTriangle[2].y = g_mod.aptTriangle[0].y + (START_STOP_Y_SIZE);
  g_mod.aptTriangle[3].x = g_mod.aptTriangle[0].x;
  g_mod.aptTriangle[3].y = g_mod.aptTriangle[0].y;
  
  g_mod.aptBreak[0].x = g_mod.aptTriangle[0].x;
  g_mod.aptBreak[0].y = g_mod.aptTriangle[0].y;
  g_mod.aptBreak[1].x = g_mod.aptTriangle[0].x;
  g_mod.aptBreak[1].y = g_mod.aptTriangle[2].y;
  g_mod.aptBreak[2].x = g_mod.aptTriangle[1].x;
  g_mod.aptBreak[2].y = g_mod.aptTriangle[0].y;
  g_mod.aptBreak[3].x = g_mod.aptTriangle[1].x;
  g_mod.aptBreak[3].y = g_mod.aptTriangle[2].y;
  
  g_mod.aptClear[0].x = g_mod.aptTriangle[0].x-2;
  g_mod.aptClear[0].y = g_mod.aptTriangle[0].y-1;
  g_mod.aptClear[1].x = g_mod.aptTriangle[1].x+3;
  g_mod.aptClear[1].y = g_mod.aptTriangle[2].y+3;

     
  *(void**)&e = (void *)SendMessage(this_mod->hwndParent,WM_WA_IPC,(LPARAM)0,IPC_GET_EMBEDIF);

  if (!e)
  {
    MessageBox(this_mod->hwndParent,"This plugin requires Winamp 5.0+","error",MB_OK);
    return 1;
  }

  parent = e(&g_mod.WindowState);

  SetWindowText(g_mod.WindowState.me, this_mod->description); // set our window title
  
  // ***** Register our window class
  {  
    WNDCLASS wc;
    memset(&wc,0,sizeof(wc));
    wc.lpfnWndProc = WndAppProc;              // our window procedure
    wc.hInstance = this_mod->hDllInstance;    // hInstance of DLL
    wc.lpszClassName = szAppName;             // our window class name
    if (!RegisterClass(&wc)) 
    {
      MessageBox(this_mod->hwndParent,"Error registering window class","error",MB_OK);
      return 1;
    }
  }

  // ***** create out application window
  styles = WS_VISIBLE|WS_CHILDWINDOW|WS_OVERLAPPED|WS_CLIPCHILDREN|WS_CLIPSIBLINGS;
  g_mod.hMainWnd = CreateWindowEx(
    0,                        // these exstyles put a nice small frame, 
                              // but also a button in the taskbar
    szAppName,                // our window class name
    NULL,                     // no title, we're a child
    styles,                   // do not make the window visible 
    cfg.config_x,cfg.config_y,// screen position (read from config)
    g_mod.width, g_mod.height,// width & height of window (need to adjust client area later)
    parent,                   // parent window (winamp main window)
    NULL,                     // no menu
    this_mod->hDllInstance,   // hInstance of DLL
    0);                       // no window creation data

  if (!g_mod.hMainWnd) 
  {
    MessageBox(this_mod->hwndParent,"Error creating window","error",MB_OK);
    return 1;
  }

  SetWindowLong(g_mod.hMainWnd,GWL_USERDATA,(LONG)this_mod); // set our reference to a "this" pointer
  SendMessage(this_mod->hwndParent, WM_WA_IPC, (int)g_mod.hMainWnd, IPC_SETVISWND);

  // ***** create our doublebuffer for paint()
  g_mod.memDC = CreateCompatibleDC(NULL);
  ZeroMemory(&bmi, sizeof bmi);
  bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi.bmiHeader.biWidth = g_mod.width;
  bmi.bmiHeader.biHeight = -g_mod.height;
  bmi.bmiHeader.biPlanes = 1;
  bmi.bmiHeader.biBitCount = 32;
  bmi.bmiHeader.biCompression = BI_RGB;
  bmi.bmiHeader.biSizeImage = 0;
  bmi.bmiHeader.biXPelsPerMeter = 0;
  bmi.bmiHeader.biYPelsPerMeter = 0;
  bmi.bmiHeader.biClrUsed = 0;
  bmi.bmiHeader.biClrImportant = 0;
  g_mod.memBM = CreateDIBSection(g_mod.memDC, &bmi, DIB_RGB_COLORS, &bits, NULL, 0);
  g_mod.oldBM = (HBITMAP)SelectObject(g_mod.memDC, g_mod.memBM);

  process_colors(this_mod, COL_INIT);

  if ( InitComPortConfig(g_mod.SioCfg, cfg.ComPort, CBR_57600) )
  {
    OpenComPort(lights.hCommPort, g_mod.SioCfg);
    if ( lights.hCommPort != INVALID_HANDLE_VALUE )
    { // ***** send JumpToConfig --> the HandHelt can be used for testing also
      sprintf_char(lights.TxBuf, ">2JC\n");
      WriteComPort(lights.hCommPort, lights.TxBuf, 5); 
    }
  }

  // ***** clear our screen with BK-color 
  {
    RECT r={0, 0, g_mod.width, g_mod.height};
    FillRect(g_mod.memDC, &r, g_mod.bk_brush);
  }

  // ***** show the window
  ShowWindow(parent,SW_SHOWNORMAL);
  return 0;
}


// ***************************************************************************
bool draw_run_stop(DWORD TickCnt)
{
  if ( g_mod.ModBreaking )
  { 
    if ( (TickCnt-g_mod.ModBreakTickCnt) >= 800 )
    {
      g_mod.ModBreakTickCnt = TickCnt;
      g_mod.ModBreakInv = !g_mod.ModBreakInv;
    }
    if ( g_mod.ModBreakInv )
    { // ***** draw inverse
      SelectObject(g_mod.memDC, g_mod.bk_brush);
      Rectangle(g_mod.memDC, g_mod.aptClear[0].x, g_mod.aptClear[0].y, g_mod.aptClear[1].x, g_mod.aptClear[1].y);
      SelectObject(g_mod.memDC, g_mod.txt_pen);
      Polyline(g_mod.memDC, &g_mod.aptBreak[0], 2);
      Polyline(g_mod.memDC, &g_mod.aptBreak[2], 2);
    }
    else
    {
      SelectObject(g_mod.memDC, g_mod.txt_brush);
      Rectangle(g_mod.memDC, g_mod.aptClear[0].x, g_mod.aptClear[0].y, g_mod.aptClear[1].x, g_mod.aptClear[1].y);
      SelectObject(g_mod.memDC, g_mod.bk_pen);
      Polyline(g_mod.memDC, &g_mod.aptBreak[0], 2);
      Polyline(g_mod.memDC, &g_mod.aptBreak[2], 2);
    }
    if ( g_mod.ModBreakUpd )
    {
      g_mod.ModBreakUpd = false;
      return true;
    }
  }
  else
  {
    SelectObject(g_mod.memDC, g_mod.txt_brush);
    Rectangle(g_mod.memDC, g_mod.aptClear[0].x, g_mod.aptClear[0].y, g_mod.aptClear[1].x, g_mod.aptClear[1].y);
    SelectObject(g_mod.memDC, g_mod.bk_pen);
    Polyline(g_mod.memDC, g_mod.aptTriangle, 4);   
  }
  return 0;
}

// ***************************************************************************
// render function for analyzer. 
// Returns 0 if successful, 1 if visualization should end.
int renderSpec(struct winampVisModule *this_mod)
{
  static TCHAR Info[50];
  DWORD TickCnt = GetTickCount();
  bool UpdateLigths = false;

  HBRUSH old = (HBRUSH)GetCurrentObject(g_mod.memDC, OBJ_BRUSH);
  int x, msk, pos_x1, pos_x2;
  // ***** pre-calculate analyzer data above half spectrum data
  int PeakDiff[LAMP_CNT], Peak[LAMP_CNT]; 
  int Idx, SpecColumn[LAMP_CNT][LAMP_LEN] = {{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0}};

  Idx = 0;
  msk = 0;
  for (x = 32; x < (576-32); x+= 4) // (512/16)/8
  {
    if ( this_mod->nCh > 1 )
    { // ***** use data from stereo channels      
      SpecColumn[0][Idx] += (this_mod->spectrumData[0][x]<<2);
      SpecColumn[1][Idx] += (this_mod->spectrumData[1][x]<<2);
    }
    else
    { // ***** use data from mono channel
      SpecColumn[0][Idx] += (this_mod->spectrumData[0][x]<<2);
      SpecColumn[1][Idx] += (this_mod->spectrumData[0][x]<<2);
    }
    if (++msk >= 10) // 512/4/8 = 16 (set smaller!!!)
    {
      msk = 0;
      // ***** combine data for upto 8 lights
      if ( ++Idx >= LAMP_LEN )
      {
        break;
      }
    }
  }

  // ***** clear background of spectrum view
  SelectObject(g_mod.memDC, g_mod.bk_brush);
  SelectObject(g_mod.memDC, g_mod.null_pen);
  // ***** clear our screen with BK-color 
  if (g_mod.ModRefresh)
  {
    g_mod.ModRefresh--;
    RECT r={0, 0, g_mod.width, g_mod.height};
    FillRect(g_mod.memDC, &r, g_mod.bk_brush);
  }
  else
  {
    Rectangle(g_mod.memDC, 0, SPEC_Y_POS, SPEC_MOD_X_SIZE, SPEC_Y_POS+SPEC_HEIGHT);
  }

  // ***** pre-calculate analyzer data
  for (x = 0, msk=128; x < LAMP_LEN; x++, msk=(msk>>1))
  { // ***** trim pre-calculateted data into view space
    SpecColumn[0][x] = (int)((double)(SpecColumn[0][x]) * cfg.SpecReduce[0][x]);
    SpecColumn[1][x] = (int)((double)(SpecColumn[1][x]) * cfg.SpecReduce[1][x]);

    if (SpecColumn[0][x] > (SPEC_HEIGHT) )
    {
      SpecColumn[0][x] = (SPEC_HEIGHT);
    }
    if (SpecColumn[1][x] > (SPEC_HEIGHT) )
    {
      SpecColumn[1][x] = (SPEC_HEIGHT);
    }
    
    lights.PeakHistory[0][x][lights.PeakIdx] = SpecColumn[0][x];
    lights.PeakHistory[1][x][lights.PeakIdx] = SpecColumn[1][x];
    // ***** draw (valid) debounced spectrum data
    if ( lights.PeakValid )
    {
      Peak[0] = lights.PeakHistory[0][x][0];
      Peak[1] = lights.PeakHistory[1][x][0];
      
      for (Idx = 1; Idx < cfg.SpecDebounceCnt; Idx ++)
      {
        Peak[0] += lights.PeakHistory[0][x][Idx];
        Peak[0] /=2;
        Peak[1] += lights.PeakHistory[1][x][Idx];
        Peak[1] /=2;
      }

      // ***** first lamps
      SelectObject(g_mod.memDC, g_mod.bars_brush[0][x]);
      pos_x1 = x*(SPEC_MOD_X_SIZE/(LAMP_LEN*LAMP_CNT)); 
      pos_x2 = pos_x1 + (SPEC_MOD_X_SIZE/(LAMP_LEN*LAMP_CNT));
      //SetBrushOrgEx (g_mod.memDC, pos_x1, (SPEC_Y_POS+SPEC_HEIGHT), &pt);
      Rectangle(g_mod.memDC,pos_x1,(SPEC_Y_POS+SPEC_HEIGHT)-Peak[0],pos_x2,(SPEC_Y_POS+SPEC_HEIGHT));
      // ***** second lamps
      SelectObject(g_mod.memDC, g_mod.bars_brush[1][x]);
      pos_x1 += (SPEC_MOD_X_SIZE/(LAMP_CNT));
      pos_x2 += (SPEC_MOD_X_SIZE/(LAMP_CNT));
      //SetBrushOrgEx (g_mod.memDC, pos_x1, (SPEC_Y_POS+SPEC_HEIGHT), &pt);
      Rectangle(g_mod.memDC,pos_x1,(SPEC_Y_POS+SPEC_HEIGHT)-Peak[1],pos_x2,(SPEC_Y_POS+SPEC_HEIGHT));

      if ( g_mod.ModBreaking )
      {
        lights.LastTickCmp = TickCnt;
      }
      else
      {
        if ( (TickCnt-lights.LastTickCmp) >= cfg.SpecRefreshTime )
        {
          for (Idx=0; Idx<LAMP_CNT; Idx++)
          {
            if ( Peak[Idx] < lights.LastCmpPeak[Idx][x] )
            {
              PeakDiff[Idx] = -(lights.LastCmpPeak[Idx][x] - Peak[Idx]);
            }
            else
            {
              PeakDiff[Idx] = Peak[Idx] - lights.LastCmpPeak[Idx][x];
            }
            if ( abs(PeakDiff[Idx]) >= cfg.SpecCmpDiff )
            {
              lights.LastCmpPeak[Idx][x] = Peak[Idx];
              if ( PeakDiff[Idx] > 0 )
              { // ***** try to set ON
                if ( !(lights.LightVal[Idx] & msk) )
                {
                  lights.LightVal[Idx] |= msk;
                  UpdateLigths = true;
                }
              }
              else
              { // ***** try to set OFF
                if ( lights.LightVal[Idx] & msk )
                { 
                  lights.LightVal[Idx] &= ((~(msk))&0xFF);
                  UpdateLigths = true;
                }
              }
            }
          }
        }
        if ( UpdateLigths )
        {
          lights.LastTickCmp = TickCnt;
        }
      }
    }
  }
  if ( ++lights.PeakIdx >= cfg.SpecDebounceCnt )
  {
    lights.PeakIdx   = 0;
    lights.PeakValid = true;
  }

  UpdateLigths |= draw_run_stop(TickCnt);
  SelectObject(g_mod.memDC, g_mod.null_pen);

  // ***** time for new update
  if ( UpdateLigths )
  {
    // ***** clear background
    SelectObject(g_mod.memDC, g_mod.bk_brush);
    Rectangle(g_mod.memDC, 0, 0, SPEC_MOD_X_SIZE, SPEC_LAMP_HEIGHT);

    // ***** send to COM port
    if ( lights.hCommPort != INVALID_HANDLE_VALUE )
    {
      //sprintf_char(lights.TxBuf, ">3L%2.2LX\n", lights.Lights);
      sprintf_char(lights.TxBuf, ">5L%2.2LX%2.2LX\n", lights.LightVal[0], lights.LightVal[1]);
      WriteComPort(lights.hCommPort, lights.TxBuf, 8); 
    }

    // ***** draw new value on screen
    msk = 0x80;
    for (Idx = 0; Idx < LAMP_LEN; Idx++, msk/=2)
    {
      if ( lights.LightVal[0] & msk ) 
      {
        SelectObject(g_mod.memDC, g_mod.lamp_brush[0][Idx]);
        Ellipse(g_mod.memDC, Idx*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (Idx+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
      }
      if ( lights.LightVal[1] & msk ) 
      {
        SelectObject(g_mod.memDC, g_mod.lamp_brush[1][Idx]);
        Ellipse(g_mod.memDC, (LAMP_LEN+Idx)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS, (LAMP_LEN+Idx+1)*SPEC_LAMP_SIZE, SPEC_LAMP_Y_POS+SPEC_LAMP_SIZE);
      }
    }
  }

  // ***** append some space for erase backgrond of older (longer) text
  if ( lights.hCommPort != INVALID_HANDLE_VALUE )
  {
    x = stprintf(Info,"Refresh: %d ms    ", TickCnt - lights.LastTickCnt);
  }
  else
  {
    if (g_mod.SioCfg.ComPort == PIPE_IDX)
    {
        x = stprintf(Info,"no Pipe !    ");
    }
    else
    {
        x = stprintf(Info,"no COM port !    ");
    }
  }


  SetBkColor(g_mod.memDC, ModColors[COL_BACK_IDX]);
  SetTextColor(g_mod.memDC, ModColors[COL_TEXT_IDX]);
  
  TextOut(g_mod.memDC, SPEC_INFO_X_POS, SPEC_INFO_Y_POS, Info, x);

  lights.LastTickCnt = TickCnt;
  SelectObject(g_mod.memDC, old);

  { // copy doublebuffer to window
    HDC hdc = GetDC(g_mod.hMainWnd);
    BitBlt(hdc, 0, 0, SPEC_MOD_X_SIZE, SPEC_MOD_Y_SIZE, g_mod.memDC, 0, 0, SRCCOPY);
    ReleaseDC(g_mod.hMainWnd, hdc);
  }
  return 0;
}


// ***************************************************************************
// render function for VU meter. 
// Returns 0 if successful, 1 if visualization should end.
int renderVu(struct winampVisModule *this_mod)
{
  static TCHAR Info[50];
  bool UpdateLigths = false;
  int PeakDiff = 0;
  int curr_peak = 0;
  int total, LoopTimeout;
  int x, len;
  DWORD TickCnt = GetTickCount(), Refresh;
  HBRUSH old = (HBRUSH)GetCurrentObject(g_mod.memDC, OBJ_BRUSH);

  // ***** calculate VU meter value and draw it as debounced value
  int last_a, last_b;
  int total_a = 0, total_b = 0;

  if (cfg.VuUsesWaveData)
  { // ***** use waveform data
    last_a  = this_mod->waveformData[0][0];
    last_b  = this_mod->waveformData[1][0];
    for (x = 1; x < 576; x +=4)
    {
      total_a += abs(last_a - (this_mod->waveformData[0][x]));
      last_a = (this_mod->waveformData[0][x]);
      total_b += abs(last_b - (this_mod->waveformData[1][x]));
      last_b = (this_mod->waveformData[1][x]);
    }
    total_a /= (576/(2*4)); 
    if (total_a > VU_MOD_X_SIZE) total_a = VU_MOD_X_SIZE;
    total_b /= (576/(2*4)); 
    if (total_b > VU_MOD_X_SIZE) total_b = VU_MOD_X_SIZE;
  }
  else
  {  // ***** use spectrum data
    for (x = 0; x < 576/2; x+=4)
    {
      total_a += this_mod->spectrumData[0][x];
      total_b += this_mod->spectrumData[1][x];
    }
    total_a /= (576/(12*4)); 
    if (total_a > VU_MOD_X_SIZE) total_a = VU_MOD_X_SIZE;
    total_b /= (576/(12*4)); 
    if (total_b > VU_MOD_X_SIZE) total_b = VU_MOD_X_SIZE;
  }

  if (this_mod->nCh > 1)
  { // ***** two stereo channels
    total = (total_a + total_b) / 2;
  }
  else
  { // ****** one mono channel
    total = total_a;
    total_b = total_a;
  }
  
  // ***** calculate debounced VU meter value
  lights.PeakHistory[0][0][lights.PeakIdx] = total_a;
  lights.PeakHistory[1][0][lights.PeakIdx] = total_b;
  if ( ++lights.PeakIdx >= cfg.SpecDebounceCnt )
  {
    lights.PeakIdx   = 0;
    lights.PeakValid = true;
  }

  int curr_peak_a, curr_peak_b;

  if ( lights.PeakValid )
  {
    curr_peak_a = 0;
    curr_peak_b = 0;
    for (x = 0; x < cfg.SpecDebounceCnt; x ++)
    {
      curr_peak_a += lights.PeakHistory[0][0][x];
      curr_peak_b += lights.PeakHistory[1][0][x];
    }
    curr_peak_a /= cfg.SpecDebounceCnt;
    curr_peak_b /= cfg.SpecDebounceCnt;
    curr_peak = (curr_peak_a + curr_peak_b) / 2;
    // ***** clear VU meter background and draw both channels
    SelectObject(g_mod.memDC, g_mod.bk_brush);
    SelectObject(g_mod.memDC, g_mod.null_pen);
    POINT pt;
    PatBlt(g_mod.memDC,0,VU_A_Y_POS, VU_MOD_X_SIZE, VU_INFO_Y_POS, PATCOPY);
    SelectObject(g_mod.memDC, g_mod.bars_brush[0][0]);
    SetBrushOrgEx (g_mod.memDC, 0, VU_A_Y_POS, &pt);
    PatBlt(g_mod.memDC,0,VU_A_Y_POS, curr_peak_a,VU_B_Y_POS-VU_A_Y_POS, PATCOPY);

    SelectObject(g_mod.memDC, g_mod.bars_brush[0][1]);
    SetBrushOrgEx (g_mod.memDC, 0, VU_B_Y_POS, &pt);
    PatBlt(g_mod.memDC,0,VU_B_Y_POS, curr_peak_b, VU_INFO_Y_POS-VU_B_Y_POS, PATCOPY);

    // ***** test VU peak for update of lights
    if ( g_mod.ModBreaking )
    {
      lights.LastTickCmp = TickCnt;
    }
    else
    {
      if ( curr_peak < lights.LastCmpPeak[0][0] )
      {
        PeakDiff = lights.LastCmpPeak[0][0] - curr_peak;
      }
      else
      {
        PeakDiff = curr_peak - lights.LastCmpPeak[0][0];
      }
      if ( (PeakDiff >= cfg.VuCmpDiff) )
      {
        if ( (TickCnt-lights.LastTickCmp) >= cfg.VuRefreshTime/2 )
        {
          UpdateLigths = true;
        }
      }
    }
  }

  // ***** time for new update
  if ( UpdateLigths )
  {
    if (++lights.ChgDiffCnt >= cfg.VuChgDiffCnt)
    {
      lights.ChgDiffCnt = 0;
    }
    else
    {
      UpdateLigths = 0;
    }
  }
  if ( UpdateLigths )
  {
    BYTE Cnt, Mask, Next0, Next1, Cmp;

    lights.LastTickCmp = TickCnt;

    // ***** create new lights and take over value on good validation
    LoopTimeout = 0;
    
    do
    {
      Next0 = lights.LightVal[0];
      Next0 ^= curr_peak;
      Next0 ^= rand()/13;              // create new value
      Cmp = lights.LightVal[0] ^ Next0;
      Cnt = 0;
      Mask = 1;
      // ***** count nr of changed bits
      for (x = 0; x < LAMP_LEN; x ++)
      {
        if ( Cmp & Mask ) Cnt++;
        Mask <<= 1;
      }
      if ( Cnt > cfg.VuChangeCnt )
      {
        Cnt = 0;
      }
      else
      {
        Cnt = 0;
        Mask = 1;
        // ***** count nr of active bits in next value
        for (x = 0; x < LAMP_LEN; x ++)
        {
          if ( Next0 & Mask ) Cnt++;
          Mask <<= 1;
        }
        if ( Cnt > cfg.VuMaxCnt )
        {
          Cnt = 0;
        }
        if ( Cnt < cfg.VuMinCnt )
        {
          Cnt = 0;
        }
      }
      Next1 = lights.LightVal[1];
      Next1 ^= curr_peak;
      Next1 ^= rand()/13;              // create new value

      // ***** avoid long looping with side effect for music thread
      if ( ++LoopTimeout > 13)
      {
        if (!Cnt)
        {
          Cnt = Next0;
        }
      }
    }while( !Cnt );

    lights.LightVal[0] = Next0;
    lights.LightVal[1] = Next1;
  }

  UpdateLigths |= draw_run_stop(TickCnt);
  SelectObject(g_mod.memDC, g_mod.null_pen);

  if ( UpdateLigths )
  {
    // ***** clear background
    SelectObject(g_mod.memDC, g_mod.bk_brush);
    Rectangle(g_mod.memDC,0,0,VU_MOD_X_SIZE,VU_LAMP_HEIGHT);

    // ***** send to COM port
    if ( lights.hCommPort != INVALID_HANDLE_VALUE )
    {
      sprintf_char(lights.TxBuf, ">5L%2.2LX%2.2LX\n", lights.LightVal[0], lights.LightVal[1]);
      WriteComPort(lights.hCommPort, lights.TxBuf, 8); 
    }

    // ***** draw new value on screen
    for (int Idx = 0; Idx < LAMP_LEN; Idx ++)
    {
      if ( lights.LightVal[0] & (1<<(7-Idx)) )
      {
        SelectObject(g_mod.memDC, g_mod.lamp_brush[0][Idx]);
        Ellipse(g_mod.memDC, Idx*VU_LAMP_SIZE, VU_LAMP_Y_POS, (Idx+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
      }
      if ( lights.LightVal[1] & (1<<(7-Idx)) )
      {
        SelectObject(g_mod.memDC, g_mod.lamp_brush[1][Idx]);
        Ellipse(g_mod.memDC, (LAMP_LEN+Idx)*VU_LAMP_SIZE, VU_LAMP_Y_POS, (LAMP_LEN+Idx+1)*VU_LAMP_SIZE, VU_LAMP_Y_POS+VU_LAMP_SIZE);
      }
    }
  }

  Refresh = TickCnt - lights.LastTickCnt;
  // ***** append some space for erase backgrond of older (longer) text
  if ( lights.hCommPort != INVALID_HANDLE_VALUE )
  {
    len = stprintf(Info,"Refresh: %d ms    ", Refresh);
  }
  else
  {
    if (g_mod.SioCfg.ComPort == PIPE_IDX)
    {
        len = stprintf(Info,"no Pipe !    ");
    }
    else
    {
        len = stprintf(Info,"no COM port !    ");
    }
  }
#ifdef MOD_TRACE 
  len = stprintf(Info,"%.50s", dbg_trace);
#endif

  SetBkColor(g_mod.memDC, ModColors[COL_BACK_IDX]);
  SetTextColor(g_mod.memDC, ModColors[COL_TEXT_IDX]);
  TextOut(g_mod.memDC, VU_INFO_X_POS, VU_INFO_Y_POS, Info, len);

  lights.LastTickCnt = TickCnt;
  SelectObject(g_mod.memDC, old);

  { // copy doublebuffer to window
    HDC hdc = GetDC(g_mod.hMainWnd);
    BitBlt(hdc, 0, 0, VU_MOD_X_SIZE, VU_MOD_Y_SIZE, g_mod.memDC, 0, 0, SRCCOPY);
    ReleaseDC(g_mod.hMainWnd, hdc);
  }
  return 0;
}


// ***************************************************************************
// cleanup (opposite of init()). 
// Destroys the window, unregisters the window class
void quit(struct winampVisModule *this_mod)
{
  SendMessage(this_mod->hwndParent, WM_WA_IPC, 0, IPC_SETVISWND);

  config_write(this_mod, true);           // write configuration (window position)
  SelectObject(g_mod.memDC,g_mod.oldBM);  // delete our doublebuffer
  DeleteObject(g_mod.memDC);
  DeleteObject(g_mod.memBM);  

  process_colors(this_mod, COL_DELETE);

  CloseComPort(lights.hCommPort);

  // ***** delete our window
  if (g_mod.WindowState.me) 
  {
    SetForegroundWindow(this_mod->hwndParent);
    DestroyWindow(g_mod.WindowState.me);
  }
  UnregisterClass(szAppName,this_mod->hDllInstance); // unregister window class
}


// ***************************************************************************
void next_preset()
{
}

// ***************************************************************************
void previous_preset()
{
}

// ***************************************************************************
void load_random_preset()
{
}

// ***************************************************************************
void set_random(bool flag)
{
  g_mod.random_presets_flag = flag;
}

// ***************************************************************************
void go_fullscreen()
{
  if ( g_mod.app_plugin )
  {
    if (SendMessage(g_mod.app_plugin->hwndParent,WM_WA_IPC,0,IPC_IS_PLAYING_VIDEO)>1) 
    {
      MessageBox(g_mod.app_plugin->hwndParent, "Can't go fullscreen while video is playing", "SVis", 0);
    }
    else
    {
      SendMessage(g_mod.app_plugin->hwndParent,WM_WA_IPC,1,IPC_SET_VIS_FS_FLAG);
      // ... now do the work of actually going fullscreen ...
    }
  }
}

// ***************************************************************************
void go_windowed()
{
  if ( g_mod.app_plugin )
  {
    SendMessage(g_mod.app_plugin->hwndParent,WM_WA_IPC,0,IPC_SET_VIS_FS_FLAG);
  }
}

// ***************************************************************************
void open_configuration()
{
}

// ***************************************************************************
void open_popup_menu()
{
}

// ***************************************************************************
// window procedure for our application window
LRESULT CALLBACK WndAppProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
    // ***************************************************************************
    case WM_PAINT:
      { // ***** update from doublebuffer
        PAINTSTRUCT ps;
        RECT r;
        HDC hdc = BeginPaint(hwnd,&ps);
        GetClientRect(hwnd,&r);
        BitBlt(hdc,0,0,r.right,r.bottom,g_mod.memDC,0,0,SRCCOPY);
        EndPaint(hwnd,&ps);
      }
      return 0;

    // ***************************************************************************
    case WM_CREATE:        
      if ( !cfg.CfgIsValid )
      { // ***** load settings from INI file
        config_read(g_mod.app_plugin);
      }
      return 0;

    // ***************************************************************************
    case WM_ERASEBKGND:   
      return 0;

    // ***************************************************************************
    case WM_LBUTTONDOWN:
      {
        POINTS pt;
        pt = MAKEPOINTS(lParam);
        if (wParam == MK_LBUTTON)
        {
          if ( (pt.x >= g_mod.aptClear[0].x) && (pt.x <= g_mod.aptClear[1].x) )
          {
            if ( (pt.y >= g_mod.aptClear[0].y) && (pt.y <= g_mod.aptClear[1].y) )
            {
              g_mod.ModBreaking = !g_mod.ModBreaking;
              g_mod.ModBreakUpd = false;
            }
          }
        }
      }
      return 0;

    // ***************************************************************************
    case WM_CHAR:
      {
        BYTE msk = 0, idx;
#ifdef MOD_TRACE 
        stprintf(dbg_trace,"char: %X,%X", wParam, lParam);
#endif
        if (g_mod.ModBreaking)
        {
          if (OemKeyScan(wParam) & 0x10000)
          {
            switch (wParam)
            {
              case 0x21: msk = 0x01; break;
              case 0x22: msk = 0x02; break;
              case 0xA7: msk = 0x04; break;
              case 0x24: msk = 0x08; break;
              case 0x25: msk = 0x10; break;
              case 0x26: msk = 0x20; break;
              case 0x2F: msk = 0x40; break;
              case 0x28: msk = 0x80; break;
              case 0x3D:
                g_mod.ModBreaking = !g_mod.ModBreaking;
                g_mod.ModBreakUpd = false;
              default:
                break;
            }
            idx = 1;
          }
          else
          {
            if ( (wParam >= 0x31) && (wParam <= 0x38) )
            {
              msk = (1<<(wParam-0x31));
              idx = 0;
            }
          }
        }
        else
        {
          if (wParam == 0x3D)
          {
            if (OemKeyScan(wParam) & 0x10000)
            {
              g_mod.ModBreaking = !g_mod.ModBreaking;
              g_mod.ModBreakUpd = false;
            }
          }
        }
        if (msk)
        {
          lights.LightVal[idx] ^= msk;
          g_mod.ModBreakUpd = true;
        }
      }
      return 0;

    // ***************************************************************************
    case WM_DESTROY: 
      PostQuitMessage(0); 
      return 0;

    // ***************************************************************************
    case WM_KEYDOWN: // pass keyboard messages to main winamp window (for processing)
    case WM_KEYUP:
      {  // get this_mod from our window's user data
        winampVisModule *this_mod = (winampVisModule *) GetWindowLong(hwnd,GWL_USERDATA);
        PostMessage(this_mod->hwndParent,message,wParam,lParam);
      }
      return 0;

    // ***************************************************************************
    case WM_WINDOWPOSCHANGING:
      {  // get config_x and config_y for configuration
        RECT r;
        GetWindowRect(g_mod.WindowState.me,&r);
        cfg.config_x = r.left;
        cfg.config_y = r.top;
      }
      return 0;

    // ***************************************************************************
    case WM_COMMAND: 
      {
        int id = LOWORD(wParam);
        switch (id) 
        {
          // user clicked on 'next' preset button
          case ID_VIS_NEXT: 
            next_preset(); 
            break;

          // user clicked on 'previous' preset button
          case ID_VIS_PREV: 
            previous_preset(); 
            break;

          // user clicked on 'random' togglebutton
          case ID_VIS_RANDOM: 
            { // determine if we're switching random on or off or if Winamp is asking us about the state of our random flag
              bool flag = HIWORD(wParam) ? 1 : 0; 
              // are we being asked about the state of our random flag ?
              if (wParam >> 16 == 0xFFFF) 
              { // tell winamp about our state
                if ( g_mod.app_plugin )
                {
                  SendMessage(g_mod.app_plugin->hwndParent, WM_WA_IPC, g_mod.random_presets_flag, IPC_CB_VISRANDOM);
                }
                break;
              }
              // changes random_preset_flag 
              set_random(flag); 
              // if we are turning random on, we should switch to a new random preset right away
              if (flag) load_random_preset();
              break;
            }
          case ID_VIS_FS: 
            go_fullscreen(); 
            break;
          case ID_VIS_CFG: 
            open_configuration(); 
            break;
          case ID_VIS_MENU: 
            open_popup_menu(); 
            break;
        } // switch (id)
        break;
      }
    // ***************************************************************************
    default:
      break;
  } // switch (message)
  return DefWindowProc(hwnd,message,wParam,lParam);
}

// ***************************************************************************
// makes a INI-file in the winamp directory named "plugin.ini"
void config_getinifn(struct winampVisModule *this_mod, TCHAR *ini_file)
{  
  TCHAR *p;
  GetModuleFileName(this_mod->hDllInstance, ini_file, MAX_PATH);
  p = ini_file+strlen(ini_file);
  while (p >= ini_file && *p != '\\') 
  {
    p--;
  }
  if (++p >= ini_file) 
  {
    *p = 0;
  }
#pragma warning(disable:4996)
  _tcscat(ini_file, "plugin.ini");
#pragma warning(default:4996)
}


// ***************************************************************************
// load configuration settings
void config_read(struct winampVisModule *this_mod)
{
  TCHAR   Str[10], Key[20], Def[10];
  double  Val;
  int     i;
  TCHAR   ini_file[MAX_PATH];

  config_getinifn(this_mod,ini_file);

  if ( this_mod->hDllInstance )
  {
    cfg.CfgIsValid = true;
  }

  // ***** screen X position and Y position, repsectively
  cfg.config_x = GetPrivateProfileInt(this_mod->description,"Screen_x", cfg.config_x,ini_file);
  cfg.config_y = GetPrivateProfileInt(this_mod->description,"Screen_y", cfg.config_y,ini_file);


  // ***** no of com port
  cfg.ComPort = GetPrivateProfileInt(this_mod->description,"ComPort", cfg.ComPort,ini_file);
  // ***** nr of dbounce loops
  cfg.SpecDebounceCnt = GetPrivateProfileInt(this_mod->description,"SpecDebounceCnt", cfg.SpecDebounceCnt,ini_file);
  cfg.VuDebounceCnt = GetPrivateProfileInt(this_mod->description,"VuDebounceCnt", cfg.VuDebounceCnt,ini_file);
  // ***** time slice for value uptate
  cfg.SpecRefreshTime = GetPrivateProfileInt(this_mod->description,"SpecRefreshTime", cfg.SpecRefreshTime,ini_file);
  cfg.VuRefreshTime = GetPrivateProfileInt(this_mod->description,"VuRefreshTime", cfg.VuRefreshTime,ini_file);
  // ***** spectrum alalyzer
  { // compare diff. for spectrum alalyzer
    cfg.SpecCmpDiff = GetPrivateProfileInt(mod_spectr.description,"SpecCmpDiff", cfg.SpecCmpDiff,ini_file);
    // reducer multiplier for spectrum analyzer
    for (i=0; i<LAMP_LEN*LAMP_CNT; i++)
    {
      stprintf(Key, "SpecReduce%d", i);
      stprintf(Def, "%1.2lf", cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN]);
      GetPrivateProfileString(mod_spectr.description,Key, Def, Str, sizeof(Str)/sizeof(TCHAR), ini_file);
      // ***** use sscanf not in a dll !!!
      Val = _ttof((TCHAR*)&Str[0]);
      cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN] = Val;
    }
  }

  // ***** VU-meter
  { // compare diff. for VU-meter
    cfg.VuCmpDiff = GetPrivateProfileInt(mod_vu.description,"VuCmpDiff", cfg.VuCmpDiff, ini_file);
    // 0: use spectr. data, 1: waveform data
    i = GetPrivateProfileInt(mod_vu.description,"VuUsesWaveData", cfg.VuUsesWaveData, ini_file);
    if (i) { cfg.VuUsesWaveData = true; } else { cfg.VuUsesWaveData = false; }
    // nr of bits allowed for changing on update of light value
    cfg.VuChangeCnt = GetPrivateProfileInt(mod_vu.description,"VuChangeCnt", cfg.VuChangeCnt, ini_file);
    // minimum of bits that must be ON
    cfg.VuMinCnt = GetPrivateProfileInt(mod_vu.description,"VuMinCnt", cfg.VuMinCnt,ini_file);
    // maximum of bits that can be ON
    cfg.VuMaxCnt = GetPrivateProfileInt(mod_vu.description,"VuMaxCnt", cfg.VuMaxCnt,ini_file);
  }
}


// ***************************************************************************
// store configuration settings
void config_write(struct winampVisModule *this_mod, bool bPos)
{
  TCHAR string[32], Key[20];
  TCHAR ini_file[MAX_PATH];
  int i;

  config_getinifn(this_mod,ini_file);

  if ( bPos )
  {
    stprintf(string,"%d", cfg.config_x);
    WritePrivateProfileString(this_mod->description,"Screen_x",string,ini_file);
    stprintf(string,"%d", cfg.config_y);
    WritePrivateProfileString(this_mod->description,"Screen_y",string,ini_file);
  }
  else
  {
    // no of com port
    stprintf(string,"%d", cfg.ComPort);
    WritePrivateProfileString(this_mod->description,"ComPort", string, ini_file);
    if ( this_mod == &mod_spectr )
    { 
      // nr of dbounce loops
      stprintf(string,"%d", cfg.SpecDebounceCnt);
      WritePrivateProfileString(this_mod->description,"SpecDebounceCnt", string, ini_file);
      // time slice for value uptate
      stprintf(string,"%d", cfg.SpecRefreshTime);
      WritePrivateProfileString(this_mod->description,"SpecRefreshTime", string, ini_file);
      // compare diff. for spectrum alalyzer
      stprintf(string,"%d", cfg.SpecCmpDiff);
      WritePrivateProfileString(this_mod->description,"SpecCmpDiff", string, ini_file);
      // reducer multiplier for spectrum analyzer
      for (i=0; i<(LAMP_CNT*LAMP_LEN); i++)
      {
        stprintf(Key, "SpecReduce%d", i);
        stprintf(string, "%1.2lf", cfg.SpecReduce[i/LAMP_LEN][i%LAMP_LEN]);
        WritePrivateProfileString(this_mod->description, Key, string, ini_file);
      }
    }

    if ( this_mod == &mod_vu )
    { 
      // nr of dbounce loops
      stprintf(string,"%d", cfg.VuDebounceCnt);
      WritePrivateProfileString(this_mod->description,"VuDebounceCnt", string, ini_file);
      // time slice for value uptate
      stprintf(string,"%d", cfg.VuRefreshTime);
      WritePrivateProfileString(this_mod->description,"VuRefreshTime", string, ini_file);
      // compare diff. for VU-meter
      stprintf(string,"%d", cfg.VuCmpDiff);
      WritePrivateProfileString(this_mod->description,"VuCmpDiff", string, ini_file);
      // 0: use spectr. data, 1: waveform data
      stprintf(string,"%d", (int)cfg.VuUsesWaveData);
      WritePrivateProfileString(this_mod->description,"VuUsesWaveData", string, ini_file);
      // nr of bits allowed for changing on update of light value
      stprintf(string,"%d", (int)cfg.VuChangeCnt);
      WritePrivateProfileString(this_mod->description,"VuChangeCnt", string, ini_file);
      // minimum of bits that must be ON
      stprintf(string,"%d", (int)cfg.VuMinCnt);
      WritePrivateProfileString(this_mod->description,"VuMinCnt", string, ini_file);
      // maximum of bits that can be ON
      stprintf(string,"%d", (int)cfg.VuMaxCnt);
      WritePrivateProfileString(this_mod->description,"VuMaxCnt", string, ini_file);
    }
  }
}
